# -*- coding: utf-8 -*-
"""AT command protocol (threaded) for Inmarsat IDP satellite messaging modems.

This module provides a threaded serial interface that sends and receives 
AT commands, decoding/abstracting typically used operations.
Based on the PySerial threaded protocol factory, using a byte reader.

"""

from base64 import b64decode, b64encode
from collections import OrderedDict
import platform
from serial import Serial, SerialException
from serial.threaded import LineReader, ReaderThread
import threading
from time import time, sleep
from typing import Callable, Tuple, Union

try:
    import queue
except ImportError:
    import Queue as queue

try:
    from .aterror import AtCrcConfigError, AtCrcError, AtException, AtTimeout, AtUnsolicited
    from crcxmodem import get_crc, validate_crc
    import constants
    import nmea
except ImportError:
    from idpmodem.crcxmodem import get_crc, validate_crc
    from idpmodem import constants
    from idpmodem import nmea

'''
class AtException(Exception):
    """Base class for AT command exceptions."""
    pass


class AtTimeout(AtException):
    """Indicates a timeout waiting for response."""
    pass


class AtCrcError(AtException):
    """Indicates a detected CRC mismatch on a response."""
    pass


class AtCrcConfigError(AtException):
    """Indicates a CRC response was received when none expected or vice versa.
    """
    pass


class AtUnsolicited(AtException):
    """Indicates unsolicited data was received from the modem."""
    pass
'''

class AtProtocol(LineReader):
    """Threaded protocol factory for the IDP Modem.
    
    Accepts only one AT command at a time.  Handles command echo 
    terminated with <cr>, verbose response framed with <cr><lf>, 
    unsolicited data terminated with <lf>, and CRC error checking.

    Attributes:
        alive (bool): True while the factory is running.
        crc (bool): Indicates if CRC error checking is enabled.
        pending_command (str): The AT command being processed.
        responses (Queue): Queued responses to be processed as a line.
        unsolicited (Queue): Unexpected data received if no pending command.
        unsolicited_callback (Callable): optional callback function for 
        unexpected data
    """

    ERROR_CODES = constants.AT_ERROR_CODES
    DEFAULT_AT_TIMEOUT = 5

    def __init__(self,
                 crc: bool = False,
                 unsolicited_callback: Callable = None,
                 default_at_timeout: int = DEFAULT_AT_TIMEOUT):
        """Initialize with CRC and optional callback

        Args:
            crc: Use CRC error checking (for long serial line).
            unsolicited_callback: Handler for non-command data.

        """
        super(AtProtocol, self).__init__()
        self.crc = crc
        self.alive = True
        self.default_at_timeout = default_at_timeout
        self.pending_command = None
        self.responses = queue.Queue()
        self.unsolicited = queue.Queue()
        self._unsolicited_thread = threading.Thread(
            target=self._run_unsolicited,
            name='at_unsolicited',
            daemon=True)
        self._unsolicited_thread.start()
        self.unsolicited_callback = unsolicited_callback
        self._lock = threading.Lock()

    def stop(self):
        """Stop the data processing thread.
        
        Aborts pending commands, if any.
        """
        self.alive = False
        self.unsolicited.put(None)
        self.responses.put('<exit>')

    def _run_unsolicited(self):
        """Process unsolicited in a separate thread.
        
        Ensures the command response thread is not blocked.

        Raises:
            AtUnsolicited: If an unexpected handling error occurs.
        
        """
        while self.alive:
            try:
                self.handle_unsolicited(self.unsolicited.get())
            except:
                raise AtUnsolicited("Unexpected error handling unsolicited data")

    def data_received(self, data: bytearray):
        """Buffer received data and create packets for handlers.

        Args:
            data: a data byte received from the serial device

        """
        self.buffer.extend(data)
        if self.pending_command is not None:
            if data == b'\r':
                #: Echo case
                if self.buffer == self.pending_command.encode() + b'\r':
                    echo = self.buffer
                    self.buffer = bytearray(b'')
                    self.handle_packet(echo)
            elif data == b'\n':
                if (self.buffer != bytearray(b'\r\n')
                    and self.buffer != bytearray(b'\n')):
                    #: Framed/multiline response, error code or CRC
                    packet = self.buffer
                    self.buffer = bytearray(b'')
                    self.handle_packet(packet)
                elif self.buffer == bytearray(b'\n'):
                    #: (Unexpected) drop any empty lines
                    self.buffer = bytearray(b'')
        else:
            if data == b'\n':
                unsolicited = self.buffer
                self.buffer = bytearray(b'')
                self.handle_packet(unsolicited)

    def handle_packet(self, packet: bytearray):
        """Processes the buffer to unicode for parser handling.

        Duplicates LineReader method for clarity in this code.

        Args:
            packet: Raw binary data buffer from the serial port.

        """
        self.handle_line(packet.decode(self.ENCODING, self.UNICODE_HANDLING))

    def handle_line(self, line: str):
        """Enqueues lines for parsing by command handler.

        Overrides LineReader class method to distinguish unsolicited
        from expected responses.
        
        Args:
            line: The unicode string received from the serial port.
        """
        if self.pending_command is not None:
            self.responses.put(line)
        else:
            if line != '\n':
                self.unsolicited.put(line)

    def handle_unsolicited(self, unsolicited: str):
        """Calls a user-defined function with the unicode string.

        Args:
            unsolicited: A unicode string terminated by <lf>.

        """
        if unsolicited is not None:
            if self.unsolicited_callback is not None:
                self.unsolicited_callback(unsolicited)
            else: 
                print('Unsolicited message: {}'.format(
                    unsolicited.replace('\r', '<cr>').replace('\n', '<lf>')))

    @staticmethod
    def _clean_response(lines: list,
                        command_time: int=None,
                        response_time: int=None) -> list:
        """Removes empty lines from response and returns with latency.
        
        Args:
            lines: A list of reponse lines.
            command_time: The timestamp the command was sent to the modem.
            response_time: The timestamp of the first line of the response.
        
        Returns:
            Tuple with stripped lines, command_latency (or None)
        """
        if command_time is not None:
            if response_time is None: response_time = time()
            command_latency = round(response_time - command_time, 3)
        else:
            command_latency = None
        for l in range(len(lines)):
            lines[l] = lines[l].strip()
        clean = list(filter(lambda line: line != '', lines))
        return (clean, command_latency)

    def _get_crc(self, command: str) -> str:
        """Returns the command with CRC.
        
        Calculates CCITT-16 checksum.

        Args:
            command: The AT command or response to calculate CRC on
        
        Returns:
            The command with CRC appended after *

        """
        return get_crc(command)
    
    @staticmethod
    def _validate_crc(lines: list, crc: str) -> bool:
        """Calculates and validates the response CRC against expected"""
        validate = ''
        for line in lines:
            validate += '{}'.format(line)
        return validate_crc(validate, crc)

    def command(
        self, command: str, timeout: int = DEFAULT_AT_TIMEOUT
    ) -> (list, int):
        """Send an AT command and wait for the response.

        Returns the response as a list.  If an error response code was
        received then 'ERROR' will be the only string in the list.

        .. todo: generalize for OK-only response, and provide error detail

        Args:
            command: The AT command
            timeout: Time to wait for response in seconds (default 5)
        
        Returns:
            The response as a list of strings
            Command latency in seconds

        Raises:
            AtCrcError if CRC does not match.
            AtCrcConfigError if CRC was returned unexpectedly.
            AtTimeout if the request timed out.

        """
        with self._lock:  # ensure that just one thread is sending commands at once
            if timeout < self.default_at_timeout:
                timeout = self.default_at_timeout
            command = get_crc(command) if self.crc else command
            self.pending_command = command
            command_sent = time()
            self.write_line(command)
            response_received = None
            lines = []
            while self.pending_command is not None:
                try:
                    line = self.responses.get(timeout=timeout)
                    content = line.strip()
                    if content == command:
                        pass   # ignore echo
                    elif content == 'OK':
                        if response_received is None:
                            response_received = time()
                        lines.append(line)
                        if not (self.crc or '%CRC=1' in self.pending_command):
                            return self._clean_response(lines,
                                                        command_sent,
                                                        response_received)
                    elif content == 'ERROR':
                        if response_received is None:
                            response_received = time()
                        lines.append(line)
                        # wait in case CRC is following
                    elif content.startswith('*'):
                        if response_received is None:
                            response_received = time()
                        if self.crc or '%CRC=1' in self.pending_command:
                            self.pending_command = None
                            crc = content.replace('*', '')
                            if self._validate_crc(lines, crc):
                                return self._clean_response(lines,
                                                            command_sent,
                                                            response_received)
                            else:
                                raise AtCrcError(
                                    'INVALID_CRC_RESPONSE {}'.format(command))
                        else:
                            raise AtCrcConfigError('UNEXPECTED_CRC_DETECTED')
                    else:
                        if response_received is None:
                            response_received = time()
                        lines.append(line)
                except queue.Empty:
                    if not response_received:
                        raise AtTimeout('TIMEOUT ({!r})'.format(command))
                    if self.crc:
                        self.crc = False
                    return self._clean_response(lines,
                                                command_sent,
                                                response_received)


class ByteReaderThread(ReaderThread):
    """Modifies the ReaderThread class to process bytes individually.
    
    This is required due to complexities of optional checksum use 
    for long serial lines.

    """
    def run(self):
        """Reader loop"""
        self.name = 'bytereader@{}'.format(self.serial.name)
        if not hasattr(self.serial, 'cancel_read'):
            self.serial.timeout = 1
        self.protocol = self.protocol_factory()
        try:
            self.protocol.connection_made(self)
        except Exception as e:
            self.alive = False
            self.protocol.connection_lost(e)
            self._connection_made.set()
            return
        error = None
        self._connection_made.set()
        data = bytearray()
        while self.alive and self.serial.is_open:
            try:
                # read all that is there or wait for one byte (blocking)
                if self.serial.in_waiting > 0:
                    data = self.serial.read()
                    self.protocol.data_received(data)
                sleep(0.001)
            except SerialException as e:
                # probably some I/O problem such as disconnected USB serial
                # adapters -> exit
                error = e
                break
            except Exception as e:
                error = e
                break
        self.alive = False
        self.protocol.connection_lost(error)
        self.protocol = None


class _AtStatistics(object):
    """A private class for tracking AT command latency in ms."""
    def __init__(self):
        self.response_times = {
            'gnss': (0, 0),
            'non_gnss': (0, 0),
        }
    
    def update(self, command, response_latency):
        """Updates the latency statistics.
        
        Args:
            command (str): The AT command submitted.
            submit_time (int): The timestamp (unix) when it was submitted.

        """
        latency = int(response_latency * 1000)
        if '%GPS' in command:
            category = 'gnss'
        else:
            category = 'non_gnss'
        category_average = self.response_times[category][0]
        category_count = self.response_times[category][1]
        if category_average == 0:
            category_average = latency
        else:
            category_average = int((category_average + latency) / 2)
        category_count += 1
        self.response_times[category] = (category_average, category_count)


class IdpModemBusy(AtException):
    pass


class IdpModem(AtProtocol):
    """A protocol factory abstracting AT commands for an IDP modem."""

    def __init__(self):
        super(IdpModem, self).__init__()
        self._at_stats = _AtStatistics()
        # TODO REMOVE self.notifications = self._notifications_dict()
        self.busy = False
    
    def connection_made(self, transport):
        """Clears the input buffer on connect.
        TODO: not tested
        May be overridden by user subclass.
        """
        super(IdpModem, self).connection_made(transport)
        self.transport.serial.reset_input_buffer()
    
    def connection_lost(self, exc):
        """Raises an exception on disconnect.
        TODO: not tested
        May be overridden by user subclass.
        """
        super(IdpModem, self).connection_lost(exc)
    
    def command(
        self, command: str, timeout:int = 5, busy_timeout:int = 30
    ) -> list:
        """Overrides the super class function to add metrics.
        
        Args:
            command (str): The AT command
            timeout (int): The command timeout (default 5 seconds)
            busy_timeout (int): Timeout (seconds) for next waiting command
        
        Returns:
            Response object.

        """
        try:
            request_time = time()
            while self.busy:
                if time() - request_time > busy_timeout:
                    raise IdpModemBusy('{}s timeout awaiting prior command: {}'
                                    .format(busy_timeout, self.pending_command))
            self.busy = True
            (response, latency) = super(IdpModem, self).command(command=command,
                                    timeout=timeout)
            self._at_stats.update(command, latency)
            if response[0] == 'ERROR':
                (reason, latency) = super(IdpModem, self).command(
                                    command='ATS80?')
                self._at_stats.update('ATS80?', latency)
                if reason[0] == 'ERROR':
                    raise AtException('Unexpected error on ATS80?')
                response.append(self.ERROR_CODES[reason[0]])
            self.busy = False
            return response
        except (AtCrcConfigError, AtCrcError, AtTimeout) as e:
            self.busy = False
            raise e

    def error_detail(self):   #TODO REMOVE REDUNDANT VS COMMAND
        """Queries the last error code.
        
        Returns:
            Reason description string or None if error on error.
        """
        response = self.command('ATS80?')
        if response is None:
            return None
        elif response[0] == 'ERROR':
            #: TODO handle CRC error? raise?
            return response[1]
        reason = self.ERROR_CODES[response[0]]
        return reason

    def config_restore_nvm(self) -> bool:
        """Sends the ATZ command to restore from non-volatile memory.
        
        Returns:
            Boolean success.
        """
        response = self.command('ATZ')
        if response[0] == 'ERROR':
            return False
        return True

    def config_restore_factory(self) -> bool:
        """Sends the AT&F command and returns True on success."""
        response = self.command('AT&F')
        if response[0] == 'ERROR':
            return False
        return True
    
    def config_report(self) -> Tuple[dict, dict]:
        """Sends the AT&V command to retrive S-register settings.
        
        Returns:
            A tuple with two dictionaries or both None if failed
            at_config with booleans crc, echo, quiet and verbose
            reg_config with S-register tags and integer values
        """
        response = self.command('AT&V')
        if response[0] == 'ERROR':
            return (None, None)
        at_config = response[1]
        s_regs = response[2]
        echo, quiet, verbose, crc = at_config.split(' ')
        at_config = {
            "crc": bool(int(crc[4])),
            "echo": bool(int(echo[1])),
            "quiet": bool(int(quiet[1])),
            "verbose": bool(int(verbose[1])),
        }
        reg_config = {}
        for reg in s_regs.split(' '):
            name, value = reg.split(':')
            reg_config[name] = int(value)
        return (at_config, reg_config)

    def config_volatile_report(self) -> Union[dict, None]:
        """Returns key S-register settings.
        
        GNSS Mode (S39), GNSS fix timeout (S41), GNSS Continuous (S55),
        GNSS Jamming Status (S56), GNSS Jamming Indicator (S57), 
        Low power Wakeup Period (S51)

        Returns:
            Dictionary of S-register values, or None if failed
            
        """
        register_list = [
            'S39',   #: GNSS Mode
            'S41',   #: GNSS Fix Timeout
            'S51',   #: Wakeup Interval
            'S55',   #: GNSS Continuous
            'S56',   #: GNSS Jamming Status
            'S57',   #: GNSS Jamming Indicator
        ]
        command = 'AT'
        for reg in register_list:
            command += '{}?'.format(reg if command == 'AT' else ' ' + reg)
        response = self.command(command)
        if response[0] == 'ERROR':
            return None
        #: else
        response.remove('OK')
        volatile_regs = {}
        for r in range(len(response)):
            volatile_regs[register_list[r]] = int(response[r])
        return volatile_regs

    def config_nvm_save(self) -> bool:
        """Sends the AT&W command and returns result."""
        response = self.command('AT&W')
        if response[0] == 'ERROR':
            return False
        return True

    def crc_enable(self, enable: bool = True) -> bool:
        """Sends the AT%CRC command and returns success flag.
        
        Args:
            enable: turn on CRC if True else turn off

        Returns:
            True if the operation succeeded else False
        """
        command = 'AT%CRC={}'.format(1 if enable else 0)
        response = self.command(command)
        if response[0] == 'ERROR':
            return False
        self.crc = enable
        return True

    def device_mobile_id(self) -> str:
        """Returns the unique Mobile ID (Inmarsat serial number).
        
        Returns:
            MobileID string or None if error.
        """
        response = self.command("AT+GSN")
        if response[0] == 'ERROR':
            return None 
        return response[0].replace('+GSN:', '').strip()

    def device_version(self) -> Tuple[str, str, str]:
        """Returns the hardware, firmware and AT versions.
        
        Returns:
            Dict with hardware, firmware, at version or all None if error.
        """
        response = self.command("AT+GMR")
        if response[0] == 'ERROR':
            return None
        versions = response[0].replace('+GMR:', '').strip()
        fw_ver, hw_ver, at_ver = versions.split(',')
        return {'hardware': hw_ver, 'firmware': fw_ver, 'at': at_ver}

    def gnss_continuous_set(self, interval: int=0, doppler: bool=True) -> bool:
        """Sets the GNSS continous mode (0 = on-demand).
        
        Args:
            interval: Seconds between GNSS refresh.
            doppler: Often required for moving assets.
        
        Returns:
            True if successful setting.
        """
        if interval < 0 or interval > 30:
            raise ValueError('GNSS continuous interval must be in range 0..30')
        response = self.command('AT%TRK={}{}'.format(
            interval, ',{}'.format(1 if doppler else 0)))
        if response[0] == 'ERROR':
            return False
        return True

    def gnss_nmea_get(self, stale_secs: int = 1, wait_secs: int = 35,
                      nmea: list = ['RMC', 'GSA', 'GGA', 'GSV']
                      ) -> Union[list, str]:
        """Returns a list of NMEA-formatted sentences from GNSS.

        Args:
            stale_secs: Maximum age of fix in seconds (1..600)
            wait_secs: Maximum time to wait for fix (1..600)

        Returns:
            List of NMEA sentences or 'ERR_TIMEOUT_OCCURRED'

        Raises:
            ValueError if parameter out of range

        """
        NMEA_SUPPORTED = ['RMC', 'GGA', 'GSA', 'GSV']
        BUFFER_SECONDS = self.default_at_timeout
        if (stale_secs not in range(1, 600+1) or
            wait_secs not in range(1, 600+1)):
            raise ValueError('stale_secs and wait_secs must be 1..600')
        sentences = ''
        for sentence in nmea:
            sentence = sentence.upper()
            if sentence in NMEA_SUPPORTED:
                if len(sentences) > 0:
                    sentences += ','
                sentences += '"{}"'.format(sentence)
            else:
                raise ValueError('Unsupported NMEA sentence: {}'
                                 .format(sentence))
        response = self.command('AT%GPS={},{},{}'
                                .format(stale_secs, wait_secs, sentences),
                                timeout=wait_secs + BUFFER_SECONDS)
        if response[0] == 'ERROR':
            return response[1]
        response.remove('OK')
        response[0] = response[0].replace('%GPS: ', '')
        return response

    def location_get(self, stale_secs: int = 1, wait_secs: int = 35):
        """Returns a location object
        """
        nmea_sentences = self.gnss_nmea_get(stale_secs, wait_secs)
        if isinstance(nmea_sentences, str):
            return None
        location = nmea.location_get(nmea_sentences)
        return location

    def message_mo_send(self,
                        data: str,
                        data_format: int = 3,
                        name: str = None,
                        priority: int = 4,
                        sin: int = None,
                        min: int = None
                        ) -> str:
        """Submits a mobile-originated message to send.
        
        Args:
            data: 
            data_format: 
            name: 
            priority: 
            sin: 
            min: 

        Returns:
            Name of the message if successful, or the error string
        """
        if name is None:
            # Use the 8 least-signficant numbers of unix timestamp as unique
            name = str(int(time()))[-8:]
        elif len(name) > 8:
            name = name[0:8]   # risk duplicates create an ERROR resposne
        response = self.command('AT%MGRT="{}",{},{}{},{},{}'.format(
                                name,
                                priority,
                                sin,
                                '.{}'.format(min) if min is not None else '',
                                data_format,
                                '"{}"'.format(data) if data_format == 1 else data))
        if response[0] == 'ERROR':
            return response[1]
        # TODO: spin thread to check status until complete/remove
        return name

    def message_mo_state(self, name: str = None) -> list:
        """Returns the message state(s) requested.
        
        If no name filter is passed in, all available messages states
        are returned.  Returns False is the request failed.

        Args:
            name: The unique message name in the modem queue

        Returns:
            State: UNAVAILABLE, TX_READY, TX_SENDING, TX_COMPLETE, TX_FAILED or None

        """
        STATES = {
            0: 'UNAVAILABLE',
            4: 'TX_READY',
            5: 'TX_SENDING',
            6: 'TX_COMPLETE',
            7: 'TX_FAILED'
        }
        
        filter = '="{}"'.format(name) if name is not None else ''
        response = self.command("AT%MGRS{}".format(filter))
        if response[0] == 'ERROR':
            return None
        # %MGRS: "<name>",<msg_no>,<priority>,<sin>,<state>,<size>,<sent_bytes>
        response.remove('OK')
        states = []
        for res in response:
            res = res.replace('%MGRS:', '').strip()
            if len(res) > 0:
                name, number, priority, sin, state, size, sent = res.split(',')
                del number
                del priority
                del sin
                states.append({'name': name,
                            'state': STATES[int(state)],
                            'size': size,
                            'sent': sent})
        return states
    
    def message_mo_cancel(self, name: str) -> bool:
        """Cancels a mobile-originated message in the Tx ready state."""
        response = self.command('AT%MGRC={}'.format(name))
        if response[0] == 'ERROR':
            return False
        return True

    def message_mo_clear(self) -> int:
        """Clears the modem transmit queue.
        
        Returns:
            Count of messages deleted, or -1 in case of error
        """
        list_response = self.command('AT%MGRL')
        if list_response[0] == 'ERROR':
            return -1
        list_response.remove('OK')
        if '%MGRL:' in list_response:
            list_response.remove('%MGRL:')
        message_count = len(list_response)
        for msg in list_response:
            name = msg.replace('%MGRL: ', '').split(',')[0]
            del_response = self.command('AT%MGRD={}C'.format(name))
            if del_response[0] == 'ERROR':
                return -1
        return message_count

    def message_mt_waiting(self) -> Union[list, None]:
        """Returns a list of received mobile-terminated message information.
        
        Returns:
            List of (name, number, priority, sin, state, length, received)

        """
        response = self.command('AT%MGFN')
        if response[0] == 'ERROR':
            return None
        response.remove('OK')
        waiting = []
        #: %MGFN: name, number, priority, sin, state, length, bytes_received
        for res in response:
            msg = res.replace('%MGFN:', '').strip()
            if msg.startswith('"FM'):
                parts = msg.split(',')
                name, number, priority, sin, state, length, received = parts
                del number   #: unused
                waiting.append({'name': name,
                                'sin': int(sin),
                                'priority': int(priority),
                                'state': int(state),
                                'length': int(length),
                                'received': int(received)})
        return waiting

    def message_mt_get(self, name: str, data_format: int = 3,
                       verbose: bool = False) -> Union[str, dict]:
        """Returns the payload of a specified mobile-terminated message.
        
        Payload is presented as a string with encoding based on data_format. 

        Args:
            name: The unique name in the modem queue e.g. FM01.01
            data_format: text=1, hex=2, base64=3 (default)

        Returns:
            The encoded data as a string

        """
        response = self.command('AT%MGFG={},{}'.format(name, data_format))
        if response is None or response[0] == 'ERROR':
            return None
        # response.remove('OK')
        #: name, number, priority, sin, state, length, data_format, data
        parts = response[0].split(',')
        sys_msg_num, sys_msg_seq = parts[1].split('.')
        msg_sin = int(parts[3])
        data_str_no_sin = parts[7]
        if data_format == constants.FORMAT_HEX:
            data = hex(msg_sin) + data_str_no_sin.lower()
        elif data_format == constants.FORMAT_B64:
            # add SIN as base64
            databytes = bytes([msg_sin]) + b64decode(data_str_no_sin)
            data = b64encode(databytes).decode('ascii')
        elif data_format == constants.FORMAT_TEXT:
            data = '\\{:02x}'.format(msg_sin) + data_str_no_sin
        try:
            message = {
                'name': parts[0],
                'system_message_number': int(sys_msg_num),
                'system_message_sequence': int(sys_msg_seq),
                'priority': int(parts[2]),
                'sin': msg_sin,
                'state': int(parts[4]),
                'length': int(parts[5]),
                'data_format': data_format,
                'data': data
            }
            return message if verbose else message['data']
        except Exception as e:
            print(e)

    def message_mt_delete(self, name: str) -> bool:
        """Marks a Return message for deletion by the modem.
        
        Args:
            name: The unique mobile-terminated name in the queue

        Returns:
            True if the operation succeeded

        """
        response = self.command('AT%MGFM="{}"'.format(name))
        if response is None or response[0] == 'ERROR':
            return False
        return True

    def event_monitor_get(self) -> Union[list, None]:
        """Returns a list of monitored/cached events.
        As a list of <class.subclass> strings which includes an asterisk
        for each new event that can be retrieved.

        Returns:
            list of strings <class.subclass[*]> or None
        """
        result = self.command('AT%EVMON')
        if result is None or result[0] == 'ERROR':
            return None
        events = result[0].replace('%EVMON: ', '').split(',')
        '''
        for i in range(len(events)):
            c, s = events[i].strip().split('.')
            if s[-1] == '*':
                s = s.replace('*', '')
                # TODO flag change for retrieval
            events[i] = (int(c), int(s))
        '''
        return events

    def event_monitor_set(self, eventlist: list) -> bool:
        """Sets trace events to monitor.

        Args:
            eventlist: list of tuples (class, subclass)

        Returns:
            True if successfully set
        """
        #: AT%EVMON{ = <c1.s1>[, <c2.s2> ..]}
        cmd = ''
        for monitor in eventlist:
            if isinstance(monitor, tuple):
                if len(cmd) > 0:
                    cmd += ','
                cmd += '{}.{}'.format(monitor[0], monitor[1])
        result = self.command('AT%EVMON={}'.format(cmd))
        if result is None or result[0] == 'ERROR':
            return False
        return True

    @staticmethod
    def _to_signed32(n):
        """Converts an integer to signed 32-bit format."""
        n = n & 0xffffffff
        return (n ^ 0x80000000) - 0x80000000

    def event_get(
        self, event: tuple, raw: bool = True
    ) -> Union[str, dict, None]:
        """Gets the cached event by class/subclass.

        Args:
            event: tuple of (class, subclass)
            raw: Returns the raw text string if True
        
        Returns:
            String if raw=True, dictionary if raw=False or None
        """
        #: AT%EVNT=c,s
        #: res %EVNT: <dataCount>,<signedBitmask>,<MTID>,<timestamp>,
        # <class>,<subclass>,<priority>,<data0>,<data1>,..,<dataN>
        if not (isinstance(event, tuple) and len(event) == 2):
            raise AtException('event_get expects (class, subclass)')
        result = self.command('AT%EVNT={},{}'.format(event[0], event[1]))
        if result is None or result[0] == 'ERROR':
            return None
        eventdata = result[0].replace('%EVNT: ', '').split(',')
        event = {
            'data_count': int(eventdata[0]),
            'signed_bitmask': bin(int(eventdata[1]))[2:],
            'mobile_id': eventdata[2],
            'timestamp': eventdata[3],
            'class': eventdata[4],
            'subclass': eventdata[5],
            'priority': eventdata[6],
            'data': eventdata[7:]
        }
        bitmask = event['signed_bitmask']
        while len(bitmask) < event['data_count']:
            bitmask = '0' + bitmask
        i = 0
        for bit in reversed(bitmask):
            #: 32-bit signed conversion redundant since response is string
            if bit == '1':
                event['data'][i] = self._to_signed32(int(event['data'][i]))
            else:
                event['data'][i] = int(event['data'][i])
            i += 1
        # TODO lookup class/subclass definitions
        return result[0] if raw else event

    @staticmethod
    def _notifications_dict(sreg_value: int = None) -> OrderedDict:
        """Returns an OrderedDictionary as an abstracted bitmask of notifications.
        
        Args:
            sreg_value: (optional) the integer value stored in S88 or S89
        
        Returns:
            ordered dictionary corresponding to bitmask
        """
        template = OrderedDict([
            (bit, False) for bit in constants.NOTIFICATION_BITMASK])
        if sreg_value is not None:
            bitmask = bin(int(sreg_value))[2:]
            if len(bitmask) > len(template):
                bitmask = bitmask[:len(template) - 1]
            while len(bitmask) < len(template):
                bitmask = '0' + bitmask
            i = 0
            for key in reversed(template):
                template[key] = True if bitmask[i] == '1' else False
                i += 1
        return template

    def notification_control_set(self, event_map: list) -> bool:
        """Sets the event notification bitmask.

        Args:
            event_map: list of dict{event_name, bool}
        
        Returns:
            True if successful.
        """
        #: ATS88=bitmask
        # TODO REMOVE old_notifications = self.notifications.copy()
        notifications_changed = False
        old_notifications = self.notification_control_get()
        if old_notifications is None:
            return False
        for event in event_map:
            if event in old_notifications:
                binary = '0b'
                for key in reversed(old_notifications):
                    bit = '1' if old_notifications[key] else '0'
                    if key == event:
                        notify = event_map[event]
                        if old_notifications[key] != notify:
                            bit = '1' if notify else '0'
                            notifications_changed = True
                            # self.notifications[key] = notify
                    binary += bit
        if notifications_changed:
            bitmask = int(binary, 2)
            result = self.command('ATS88={}'.format(bitmask))
            if result is None or result[0] == 'ERROR':
                return False
        return True
    
    def notification_control_get(self) -> Union[OrderedDict, None]:
        """Returns the current notification configuration bitmask."""
        #: ATS88?
        result = self.command('ATS88?')
        if result is None or result[0] == 'ERROR':
            return None
        return self._notifications_dict(int(result[0]))

    def notification_check(self) -> OrderedDict:
        """Returns the current active event notification bitmask.
        Clears the value of S89 upon reading.
        """
        #: ATS89?
        result = self.command('ATS89?')
        if result is None or result[0] == 'ERROR':
            return None
        template = self._notifications_dict(int(result[0]))
        return template

    def sat_status_snr(self) -> Tuple[str, float]:
        """Returns the control state and C/No.
        
        Returns:
            Tuple with (state: int, C/No: float) or None if error.
        """
        response = self.command("ATS90=3 S91=1 S92=1 S122? S116?")
        if response is None or response[0] == 'ERROR':
            return (None, None)
        response.remove('OK')
        ctrl_state, cn_0 = response
        ctrl_state = int(ctrl_state)
        cn_0 = int(cn_0) / 100.0
        return (ctrl_state, cn_0)

    def sat_status_name(self, ctrl_state: int) -> str:
        """Returns human-readable definition of a control state value.
        
        Raises:
            ValueError if ctrl_state is not found.
        """
        for s in constants.CONTROL_STATES:
            if int(s) == ctrl_state:
                return constants.CONTROL_STATES[s]
        raise ValueError('Control state {} not found'.format(ctrl_state))

    def shutdown(self) -> bool:
        """Tell the modem to prepare for power-down."""
        response = self.command('AT%OFF')
        if response is None or response[0] == 'ERROR':
            return False
        return True

    def utc_time(self) -> Union[str, None]:
        """Returns current UTC time of the modem in ISO format."""
        response = self.command('AT%UTC')
        if response is None or response[0] == 'ERROR':
            return None
        return response[0].replace('%UTC: ', '').replace(' ', 'T') + 'Z'

    def s_register_get(self, register: str) -> Union[int, None]:
        """Returns the value of the S-register requested.

        Args:
            register: The register name/number (e.g. S80)

        Returns:
            integer value or None
        """
        if not register.startswith('S'):
            # TODO: better Exception handling
            raise Exception('Invalid S-register {}'.format(register))
        response = self.command('AT{}?'.format(register))
        if response is None or response[0] == 'ERROR':
            return None
        return int(response[0])

    def sreg_get_all(self) -> Union[list, None]:
        """Returns a list of S-register definitions.
        R=read-only, S=signed, V=volatile
        
        Returns:
            tuple(register, RSV, current, default, minimum, maximum) or None
        """
        #: AT%SREG
        #: Sreg, RSV, CurrentVal, DefaultVal, MinimumVal, MaximumVal
        result = self.command('AT%SREG')
        if result is None or result[0] == 'ERROR':
            return None
        result.remove('OK')
        reg_defs = result[2:]
        registers = []
        for row in reg_defs:
            reg_def = row.split(' ')
            reg_def = tuple(filter(None, reg_def))
            registers.append(reg_def)
        return registers

    def raw_command(self, command: str = 'AT', timeout: int = 5) -> list:
        """Sends a command and returns a list of responses and response code."""
        response = self.command(command, timeout)
        return response


def get_modem_thread(port: str = '/dev/ttyUSB0', baudrate: int = 9600) -> tuple:
    """Gets a threaded IDP modem connection / protocol factory.

    Args:
        port: The serial port name (default /dev/ttyUSB0)
        baudrate: The serial baud rate (default 9600)

    Returns:
        (IdpModem, Thread)
    """
    ser = Serial(port, baudrate=baudrate)
    t = ByteReaderThread(ser, IdpModem)
    t.start()
    transport, idp_modem = t.connect()
    del transport   #: unusued
    if platform.system() == 'Windows':
        idp_modem.default_at_timeout = 10
    return (idp_modem, t)


# Self-test
if __name__ == '__main__':
    try:
        idp_modem, t = get_modem_thread()
        try:
            connected = idp_modem.config_restore_nvm()
            if not connected:
                raise Exception('Could not connect to modem')
            print('Modem connected')
        except AtCrcConfigError:
            idp_modem.crc = True
            idp_modem.config_restore_nvm()
        crc_enabled = idp_modem.crc_enable()
        mobile_id = idp_modem.device_mobile_id()
        print('Mobile ID: {}'.format(mobile_id))
        versions = idp_modem.device_version()
        print('Versions: FW={} HW={} AT={}'.format(
              versions['firmware'], versions['hardware'], versions['at']))
        
        def send_another_command():
            sleep(0.5)
            print('Requesting satellite status')
            state = None
            snr = None
            while state is None:
                (state, snr) = idp_modem.sat_status_snr()
            print('State: {} | SNR: {} dB'.format(
                idp_modem.sat_status_name(state), snr))
            waiting = idp_modem.message_mt_waiting()
            if isinstance(waiting, list) and len(waiting) > 0:
                print('{} MT messages waiting'.format(len(waiting)))

        test_thread = threading.Thread(target=send_another_command,
                                       name='send_another_command',
                                       daemon=True)
        # test_thread.start()
        '''
        gnss_timeout = 15   # seconds
        print('Requesting GNSS location with timeout {}s'.format(gnss_timeout))
        nmea_sentences = idp_modem.gnss_nmea_get(wait_secs=gnss_timeout)
        if isinstance(nmea_sentences, list):
            for sentence in nmea_sentences:
                print('{}'.format(sentence))
        else:
            print(nmea_sentences)
        registers = idp_modem.sreg_get_all()
        print(registers)
        '''
        to_monitor = (3, 1)
        mon = idp_modem.event_monitor_set([to_monitor])
        monitored = idp_modem.event_monitor_get()
        print('Monitored: {}'.format(monitored))
        if monitored is not None:
            for mon in monitored:
                if mon[-1] == '*':
                    to_get = tuple(int(i) for i in mon[0:-1].split('.'))
                    print(idp_modem.event_get(to_get))
        idp_modem.notification_control_set({'event_cached': True})
        event_notifications = idp_modem.notification_control_get()
        print('Notifications enabled: {}'.format(event_notifications))
        event_check = idp_modem.notification_check()
        print('Notifications active: {}'.format(event_check))
        print('Latency Statistics:')
        for stat in idp_modem._at_stats.response_times:
            print('  {}: {} ms'.format(
                stat, idp_modem._at_stats.response_times[stat][0]))
        delay = 5
        print('Waiting for unsoliticted input {} seconds'.format(delay))
        sleep(delay)
    
    except KeyboardInterrupt:
        print('Interrupted by user')

    except Exception as e:
        idp_modem.stop()
        print('EXCEPTION: {}'.format(e))
    
    finally:
        t.close()



# -*- coding: utf-8 -*-
"""IsatData Pro modem constants.

This module provides mapping of constants used within an IDP modem.

"""

from collections import OrderedDict

# Message Priorities and Data Formats
PRIORITY_MT, PRIORITY_HIGH, PRIORITY_MEDH, PRIORITY_MEDL, PRIORITY_LOW = (
    0, 1, 2, 3, 4)
FORMAT_TEXT, FORMAT_HEX, FORMAT_B64 = (1, 2, 3)
# Message States
UNAVAILABLE = 0
RX_COMPLETE = 2
RX_RETRIEVED = 3
TX_READY = 4
TX_SENDING = 5
TX_COMPLETE = 6
TX_FAILED = 7
# Wakeup Intervals
WAKEUP_5_SEC = 0
WAKEUP_30_SEC = 1
WAKEUP_1_MIN = 2
WAKEUP_3_MIN = 3
WAKEUP_10_MIN = 4
WAKEUP_30_MIN = 5
WAKEUP_60_MIN = 6
WAKEUP_2_MIN = 7
WAKEUP_5_MIN = 8
WAKEUP_15_MIN = 9
WAKEUP_20_MIN = 10
WAKEUP_INTERVALS = (0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10)
# Power Mode settings (S50)
POWER_MODE_MOBILE_POWERED = 0
POWER_MODE_FIXED_POWERED = 1
POWER_MODE_MOBILE_BATTERY = 2
POWER_MODE_FIXED_BATTERY = 3
POWER_MODE_MOBILE_MINIMAL = 4
POWER_MODE_MOBILE_STATIONARY = 5
POWER_MODES = (0, 1, 2, 3, 4, 5)
# GNSS Modes
GNSS_MODES = (0, 1, 2, 10, 11, 12)
GNSS_MODE_GPS = 0
GNSS_MODE_GLONASS = 1
GNSS_MODE_BEIDOU = 2
GNSS_MODE_GPS_GLONASS = 10
GNSS_MODE_GPS_BEIDOU = 11
GNSS_MODE_GLONASS_BEIDOU = 12
# GNSS Dynamic Platform Models
GNSS_DPM_MODES = (0, 2, 3, 4, 5, 6, 7, 8)
GNSS_DPM_PORTABLE = 0
GNSS_DPM_STATIONARY = 2
GNSS_DPM_PEDESTRIAN = 3
GNSS_DPM_AUTOMOTIVE = 4
GNSS_DPM_SEA = 5
GNSS_DPM_AIR_1G = 6
GNSS_DPM_AIR_2G = 7
GNSS_DPM_AIR_4G = 8

CONTROL_STATES = {
    0: 'Stopped',
    1: 'Waiting for GNSS fix',
    2: 'Starting search',
    3: 'Beam search',
    4: 'Beam found',
    5: 'Beam acquired',
    6: 'Beam switch in progress',
    7: 'Registration in progress',
    8: 'Receive only',
    9: 'Downloading Bulletin Board',
    10: 'Active',
    11: 'Blocked',
    12: 'Confirm previously registered beam',
    13: 'Confirm requested beam',
    14: 'Connect to confirmed beam'
}

BEAMSEARCH_STATES = {
    0: 'Idle',
    1: 'Search for any traffic channel',
    2: 'Search for last acquired traffic channel',
    3: 'reserved',
    4: 'Search for another traffic channel',
    5: 'Search for bulletin board',
    6: 'Delay traffic channel search'
}

MESSAGE_STATES = {
    0: 'UNAVAILABLE',
    1: 'RX_PENDING',
    2: 'RX_COMPLETE',
    3: 'RX_RETRIEVED',
    4: 'TX_READY',
    5: 'TX_SENDING',
    6: 'TX_COMPLETE',
    7: 'TX_FAILED',
    8: 'TX_CANCELLED'
}

AT_ERROR_CODES = {
    0: 'OK',
    4: 'ERROR',
    100: 'ERR_INVALID_CRC_SEQUENCE',
    101: 'ERR_UNKNOWN_COMMAND',
    102: 'ERR_INVALID_COMMAND_PARAMETERS',
    103: 'ERR_MESSAGE_LENGTH_EXCEEDS_FORMAT_SIZE',
    104: 'ERR_RESERVED_104',
    105: 'ERR_SYSTEM_ERROR',
    106: 'ERR_QUEUE_INSUFFICIENT_RESOURCES',
    107: 'ERR_MESSAGE_NAME_ALREADY_IN_USE',
    108: 'ERR_TIMEOUT_OCCURRED',
    109: 'ERR_UNAVAILABLE',
    110: 'ERR_RESERVED_110',
    111: 'ERR_RESERVED_111',
    112: 'ERR_ATTEMPT_TO_WRITE_READ_ONLY_PARAMETER'
}

WAKEUP_PERIODS = {
    0: 'SECONDS_5',
    1: 'SECONDS_30',
    2: 'MINUTES_1',
    3: 'MINUTES_3',
    4: 'MINUTES_10',
    5: 'MINUTES_30',
    6: 'MINUTES_60',
    7: 'MINUTES_2',
    8: 'MINUTES_5',
    9: 'MINUTES_15',
    10: 'MINUTES_20'
}

POWER_MODES = {
    0: 'MOBILE_POWERED',
    1: 'FIXED_POWERED',
    2: 'MOBILE_BATTERY',
    3: 'FIXED_BATTERY',
    4: 'MOBILE_MINIMAL',
    5: 'MOBILE_PARKED'
}

GNSS_MODES = {
    'GPS': 0,               # HW v4
    'GLONASS': 1,           # HW v5
    'BEIDOU': 2,            # HW v5.2
    'GPS+GLONASS': 10,      # UBX-M80xx
    'GPS+BEIDOU': 11,       # UBX-M80xx
    'GLONASS+BEIDOU': 12    # UBX-M80xx
}

GNSS_DPM_MODES = {
    0: 'PORTABLE',
    2: 'STATIONARY',
    3: 'PEDESTRIAN',
    4: 'AUTOMOTIVE',
    5: 'SEA',
    6: 'AIR_1G',
    7: 'AIR_2G',
    8: 'AIR_4G'
}

NOTIFICATION_BITMASK = (
    'gnss_fix_new',
    'message_mt_received',
    'message_mo_complete',
    'network_registered',
    'modem_reset',
    'jamming_antenna_change',
    'modem_reset_pending',
    'wakeup_period_changed',
    'utc_time_set',
    'gnss_fix_timeout',
    'event_cached',
    'network_ping_acknowledged'
)

NOTIFICATION_BITMASK_2 = {
    0: 'GNSS_NEW_FIX',
    1: 'MESSAGE_MT_RECEIVED',
    2: 'MESSAGE_MO_COMPLETE',
    3: 'NETWORK_REGISTERED',
    4: 'MODEM_RESET',
    5: 'JAMMING_ANTENNA_CHANGE',
    6: 'MODEM_RESET_PENDING',
    7: 'WAKEUP_PERIOD_CHANGED',
    8: 'TIME_UTC_SYNC',
    9: 'GNSS_TIMEOUT',
    10: 'TRACE_EVENT_CACHED',
    11: 'NETWORK_PING_ACKNOWLEDGED'
}

TRANSMIT_STATUS = {
    4: 'RX_ONLY_NOT_REGISTERED',
    5: 'OK',
    6: 'SUSPENDED',
    7: 'MUTED',
    8: 'BLOCKED'
}

SATELLITE_GENERAL_TRACE = (
    {'subframe_number': 'uint'},
    {'traffic_channel_id': 'uint'},
    {'configuration_id': 'uint'},
    {'beam_number': 'uint'},
    {'reserved04': 'uint'},
    {'reserved05': 'uint'},
    {'operator_tx_access': 'uint'},
    {'user_tx_mute': 'uint'},
    {'tx_suspend_flags': {0: 'beam_registration', 1: 'beam_switch', 2: 'reserved', 3: 'blocked'}},
    {'active_tx_messages': 'uint'},
    {'total_tx_messages': 'uint'},
    {'tx_state': ('active', 'suspending', 'suspended')},
    {'active_rx_messages': 'uint'},
    {'beamswitch_averaging_window': 'uint'},
    {'beamswitch_averaging_count': 'uint'},
    {'c_n_x100': 'uint'},
    {'beamsample_threshold': 'uint'},
    {'beamsample_timer': 'uint'},
    {'flags': {
        0: 'registered',
        1: 'sending_beam_registration',
        4: 'beam_search',
        5: 'need_beam_sample',
        6: 'beam_switch_pending',
        8: 'gnss_valid',
        9: 'need_gnss',
        10: 'requested_gnss'
    }},
    {'gnss_state_timer': 'uint'},
    {'reserved21': 'uint'},
    {'satellite_control_state': (
        'stopped',
        'gnss_wait',
        'beam_search_start',
        'beam_search',
        'beam_found',
        'beam_acquired',
        'beam_switch_pending',
        'registration_pending',
        'receive_only',
        'bulletinboard_receive',
        'active',
        'blocked',
        'confirm_previous_registered_beam',
        'confirm_requested_beam',
        'connect_confirmed_beam'
    )},
    {'beam_search_state': (
        'idle',
        'search_any_traffic',
        'search_last_traffic',
        'reserved',
        'search_other_traffic',
        'search_bulletinboard',
        'delay_search_traffic'
    )}
)


"""IDP modem S-register definitions

IDP modem S-register definitions useful for creating a digital twin of the modem
Updated register map of default and configured values can be read from the modem
using AT%SREG which returns a human-readable table
"""

class SRegisters(object):
    """
    A private class to manage twin of the modem's S registers

    :param object: [description]
    :type object: [type]
    :return: [description]
    :rtype: [type]
    """
    # Tuples: (name[0], default[1], read-only[2], range[3], description[4], note[5])
    register_definitions = [
        ('S0', 0, True, [0, 255], 'auto answer', 'unused'),
        ('S3', 13, False, [1, 127], 'command termination character', None),
        ('S4', 10, False, [0, 127], 'response formatting character', None),
        ('S5', 8, False, [0, 127], 'command line editing character', None),
        ('S6', 0, True, [0, 255], 'pause before dial', 'unused'),
        ('S7', 0, True, [0, 255],
            'connection completion timeout', 'unused'),
        ('S8', 0, True, [0, 255], 'commia dial modifier time', 'unused'),
        ('S10', 0, True, [0, 255], 'automatic discovery delay', 'unused'),
        ('S31', 80, False, [10, 250], 'DOP threshold (x10)', None),
        ('S32', 25, False, [1, 1000],
            'position accuracy threshold [m]', None),
        ('S33', 0, False, [0, 8], 'default dynamic platform model', None),
        ('S34', 7, True, [0, 255],
            'Doppler dynamic platform model', 'Reserved'),
        ('S35', 0, False, [0, 255], 'static hold threshold [cm/s]', None),
        ('S36', 0, False, [-1, 480], 'standby timeout [min]', None),
        ('S37', 200, False, [1, 1000], 'speed accuracy threshold', None),
        ('S38', 1, True, [0, 0], 'reserved', None),
        ('S39', 0, False, [0, 2], 'GNSS mode', None),
        ('S40', 0, False, [0, 60],
            'GNSS signal satellite detection timeout', None),
        ('S41', 180, False, [60, 1200], 'GNSS fix timeout', None),
        ('S42', 65535, False, [0, 65535],
            'GNSS augmentation systems', 'Query fails'),
        ('S50', 0, False, [0, 9], 'power mode', None),
        ('S51', 0, False, [0, 6], 'wakeup interval', None),
        ('S52', 2500, True, [0, 2500], 'reserved', 'undocumented'),
        ('S53', 0, True, [0, 255], 'satcom control', None),
        ('S54', 0, True, [0, 0], 'satcom status', None),
        ('S55', 0, False, [0, 30], 'GNSS continuous mode', None),
        ('S56', 0, True, [0, 255], 'GNSS jamming status', None),
        ('S57', 0, True, [0, 255], 'GNSS jamming indicator', None),
        ('S60', 1, False, [0, 1], 'Echo', None),
        ('S61', 0, False, [0, 1], 'Quiet', None),
        ('S62', 1, False, [0, 1], 'Verbose', None),
        ('S63', 0, False, [0, 1], 'CRC', None),
        ('S64', 42, False, [0, 255],
            'prefix character of CRC sequence', None),
        ('S70', 0, True, [0, 0], 'reserved', 'undocumented'),
        ('S71', 0, True, [0, 0], 'reserved', 'undocumented'),
        ('S80', 0, True, [0, 255], 'last error code', None),
        ('S81', 0, True, [0, 255], 'most recent result code', None),
        ('S85', 22, True, [0, 0], 'temperature', None),
        ('S88', 0, False, [0, 65535], 'event notification control', None),
        # ('S89', 0, False, [0, 65535], 'event notification status', None),
        ('S90', 0, False, [0, 7], 'capture trace define - class', None),
        ('S91', 0, False, [0, 31],
            'capture trace define - subclass', None),
        ('S92', 0, False, [0, 255],
            'capture trace define - initiate', None),
        ('S93', 0, True, [0, 255],
            'captured trace property - data size', None),
        ('S94', 0, True, [0, 255],
            'captured trace property - signed indicator', None),
        ('S95', 0, True, [0, 255],
            'captured trace property - mobile ID', None),
        ('S96', 0, True, [0, 255],
            'captured trace property - timestamp', None),
        ('S97', 0, True, [0, 255],
            'captured trace property - class', None),
        ('S98', 0, True, [0, 255],
            'captured trace property - subclass', None),
        ('S99', 0, True, [0, 255],
            'captured trace property - severity', None),
        ('S100', 0, True, [0, 255], 'captured trace data 0', None),
        ('S101', 0, True, [0, 255], 'captured trace data 1', None),
        ('S102', 0, True, [0, 255], 'captured trace data 2', None),
        ('S103', 0, True, [0, 255], 'captured trace data 3', None),
        ('S104', 0, True, [0, 255], 'captured trace data 4', None),
        ('S105', 0, True, [0, 255], 'captured trace data 5', None),
        ('S106', 0, True, [0, 255], 'captured trace data 6', None),
        ('S107', 0, True, [0, 255], 'captured trace data 7', None),
        ('S108', 0, True, [0, 255], 'captured trace data 8', None),
        ('S109', 0, True, [0, 255], 'captured trace data 9', None),
        ('S110', 0, True, [0, 255], 'captured trace data 10', None),
        ('S111', 0, True, [0, 255], 'captured trace data 11', None),
        ('S112', 0, True, [0, 255], 'captured trace data 12', None),
        ('S113', 0, True, [0, 255], 'captured trace data 13', None),
        ('S114', 0, True, [0, 255], 'captured trace data 14', None),
        ('S115', 0, True, [0, 255], 'captured trace data 15', None),
        ('S116', 0, True, [0, 255], 'captured trace data 16', None),
        ('S117', 0, True, [0, 255], 'captured trace data 17', None),
        ('S118', 0, True, [0, 255], 'captured trace data 18', None),
        ('S119', 0, True, [0, 255], 'captured trace data 19', None),
        ('S120', 0, True, [0, 255], 'captured trace data 20', None),
        ('S121', 0, True, [0, 255], 'captured trace data 21', None),
        ('S122', 0, True, [0, 255], 'captured trace data 22', None),
        ('S123', 0, True, [0, 255], 'captured trace data 23', None),
    ]

    class SRegister(object):
        """
        Twin of a modem S register
        """

        def __init__(self, name, default, read_only, low, high, description, note=None):
            """
            Initializes an S register twin

            :param name: (string) name of the S register e.g. 'S50'
            :param default: (int) default value of the register
            :param read_only: (Boolean)
            :param low: (int) lowest value allowed
            :param high: (int) highest value allowed
            :param description: (string)
            :param note: (string), defaults to None
            """
            self.name = name
            self.default = default
            self.value = default
            self.read_only = read_only
            self.rng = range(low, high+1)
            self.description = description
            self.note = note

        def get(self):
            return self.value

        def set(self, value):
            error = None
            if not self.read_only:
                if value in self.rng:
                    self.value = value
                else:
                    error = "Attempt to set {} out of range.".format(
                        self.name)
            else:
                error = "Attempt to write read-only register {}".format(
                    self.name)
            return error if error is not None else value

        def read(self):
            pass

    def __init__(self):
        self.s_registers = []
        for tup in self.register_definitions:
            reg = self.SRegister(name=tup[0], default=tup[1], read_only=tup[2],
                                 low=tup[3][0], high=tup[3][1],
                                 description=tup[4], note=tup[5])
            self.s_registers.append(reg)

    def register(self, name):
        """
        Returns the register object based on its name

        :param name: (string) name of the register e.g. 'S50'
        :return: the register object
        :rtype: (SRegister)
        """
        for reg in self.s_registers:
            if reg.name == name:
                return reg




"""Helpers for operating a headless device for example Raspberry Pi.

"""

import inspect
import logging
from logging import Logger, INFO
from logging.handlers import RotatingFileHandler
from threading import Thread, Event
from time import gmtime, time
from typing import Callable, Tuple, Union

import serial.tools.list_ports as list_ports
import queue


class RepeatingTimer(Thread):
    """A repeating thread to call a function, can be stopped/restarted/changed.
    
    Embedded tasks can use threading for continuous repeated operations.  A
    *RepeatingTimer* can be started, stopped, restarted and reconfigured.

    A Thread that counts down seconds using sleep increments, then calls back 
    to a function with any provided arguments.
    Optional auto_start feature starts the thread and the timer, in this case 
    the user doesn't need to explicitly start() then start_timer().

    Attributes:
        name (str): An optional descriptive name for the Thread.
        interval (int): Repeating timer interval in seconds (0=disabled).
        sleep_chunk (float): The fraction of seconds between processing ticks.

    """
    def __init__(self,
                 seconds: int,
                 target: Callable,
                 args: Tuple = (),
                 kwargs: dict = {},
                 name: str = None,
                 logger: Logger = None,
                 log_level: int = INFO,
                 sleep_chunk: float = 0.25,
                 max_drift: int = None,
                 auto_start: bool = False,
                 defer: bool = True,
                 debug: bool = False,
                 daemon = True):
        """Sets up a RepeatingTimer thread.

        Args:
            seconds: Interval for timer repeat.
            target: The function to execute each timer expiry.
            args: Positional arguments required by the target.
            name: Optional thread name.
            logger: Optional external logger to use.
            sleep_chunk: Tick seconds between expiry checks.
            max_drift: Number of seconds clock drift to tolerate.
            auto_start: Starts the thread and timer when created.
            defer: Set if first target waits for timer expiry.
            verbose_debug: verbose logging of tick count
            kwargs: Optional keyword arguments to pass into the target.

        Raises:
            ValueError if seconds is not an integer.
        """
        if not (isinstance(seconds, int) and seconds >= 0):
            err_str = 'RepeatingTimer seconds must be integer >= 0'
            raise ValueError(err_str)
        super().__init__(daemon=daemon)
        self.name = name or '{}_timer_thread'.format(str(target))
        self._log = logger or get_wrapping_logger(name=self.name,
                                                  log_level=log_level)
        self.interval = seconds
        if target is None:
            self._log.warning('No target specified for RepeatingTimer {}'
                              .format(self.name))
        self.target = target
        self._exception = None
        self.args = args
        self.kwargs = kwargs
        self.sleep_chunk = sleep_chunk
        self._defer = defer
        self._debug = debug
        self._terminate_event = Event()
        self._start_event = Event()
        self._reset_event = Event()
        self._count = self.interval / self.sleep_chunk
        self._timesync = time()
        self.max_drift = max_drift
        if auto_start:
            self.start()
            self.start_timer()

    @property
    def sleep_chunk(self):
        return self._sleep_chunk

    @sleep_chunk.setter
    def sleep_chunk(self, value: float):
        if 1 % value != 0:
            raise ValueError('1 must be a multiple of sleep_chunk')
        self._sleep_chunk = value

    def _resync(self, max_drift: int = None) -> int:
        """Used to adjust the next countdown to account for drift.
        
        Untested.
        """
        if max_drift is not None:
            drift = time() - self._timesync % self.interval
            max_drift = 0 if max_drift < 1 else max_drift
            if drift > max_drift:
                self._log.warning('Detected drift of {}s'.format(drift))
                return drift
        return 0

    def run(self):
        """*Note: runs automatically, not meant to be called explicitly.*
        
        Counts down the interval, checking every ``sleep_chunk`` for expiry.
        """
        while not self._terminate_event.is_set():
            while (self._count > 0
                   and self._start_event.is_set()
                   and self.interval > 0):
                if self._debug:
                    if (self._count * self.sleep_chunk
                        - int(self._count * self.sleep_chunk)
                        == 0.0):
                        #: log debug message at reasonable interval
                        self._log.debug('{} countdown: {} ({}s @ step {})'
                                        .format(self.name,
                                        self._count,
                                        self.interval,
                                        self.sleep_chunk))
                if self._reset_event.wait(self.sleep_chunk):
                    self._reset_event.clear()
                    self._count = self.interval / self.sleep_chunk
                self._count -= 1
                if self._count <= 0:
                    try:
                        self.target(*self.args, **self.kwargs)
                        drift_adjust = (self.interval
                                        - self._resync(self.max_drift))
                        self._count = drift_adjust / self.sleep_chunk
                    except BaseException as e:
                        self._exception = e

    def start_timer(self):
        """Initially start the repeating timer."""
        self._timesync = time()
        if not self._defer and self.interval > 0:
            self.target(*self.args, **self.kwargs)
        self._start_event.set()
        if self.interval > 0:
            self._log.info('{} timer started ({} seconds)'.format(
                           self.name, self.interval))
        else:
            self._log.warning('{} timer will not trigger (interval 0)'.format(
                              self.name))

    def stop_timer(self):
        """Stop the repeating timer."""
        self._start_event.clear()
        self._log.info('{} timer stopped ({} seconds)'
                       .format(self.name, self.interval))
        self._count = self.interval / self.sleep_chunk

    def restart_timer(self):
        """Restart the repeating timer (after an interval change)."""
        if not self._defer and self.interval > 0:
            self.target(*self.args, **self.kwargs)
        if self._start_event.is_set():
            self._reset_event.set()
        else:
            self._start_event.set()
        if self.interval > 0:
            self._log.info('{} timer restarted ({} seconds)'.format(
                           self.name, self.interval))
        else:
            self._log.warning('{} timer will not trigger (interval 0)'.format(
                              self.name))

    def change_interval(self, seconds: int):
        """Change the timer interval and restart it.
        
        Args:
            seconds (int): The new interval in seconds.
        
        Raises:
            ValueError if seconds is not an integer.

        """
        if (isinstance(seconds, int) and seconds >= 0):
            self._log.info('{} timer interval changed (old:{} s new:{} s)'
                           .format(self.name, self.interval, seconds))
            self.interval = seconds
            self._count = self.interval / self.sleep_chunk
            self.restart_timer()
        else:
            err_str = 'RepeatingTimer seconds must be integer >= 0'
            self._log.error(err_str)
            raise ValueError(err_str)

    def terminate(self):
        """Terminate the timer. (Cannot be restarted)"""
        self.stop_timer()
        self._terminate_event.set()
        self._log.info('{} timer terminated'.format(self.name))
    
    def join(self):
        super(RepeatingTimer, self).join()
        if self._exception:
            raise self._exception
        return self.target


class SearchableQueue(object):
    """Mimics relevant FIFO queue functions to avoid duplicate commands.
    
    Makes use of queue Exceptions to mimic a standard queue.

    Attributes:
        max_size: The maximum queue depth.
    """
    def __init__(self, max_size=100):
        self._queue = []
        self.max_size = max_size
    
    def contains(self, item):
        """Returns true if the queue contains the item."""
        for i in self._queue:
            if i == item: return True
        return False

    def put(self, item, index=None):
        """Adds the item to the queue.
        
        Args:
            item: The object to add to the queue.
            index: The queue position (None=end)
        """
        if len(self._queue) > self.max_size:
            raise queue.Full
        if index is None:
            self._queue.append(item)
        else:
            self._queue.insert(index, item)

    def put_exclusive(self, item):
        """Adds the item to the queue only if unique in the queue.
        
        Args:
            item: The object to add to the queue.
        
        Raises:
            queue.Full if a duplicate item is in the queue.
        """
        if not self.contains(item):
            self.put(item)
        else:
            raise queue.Full('Duplicate item in queue')

    def get(self):
        """Pops the first item from the queue.
        
        Returns:
            An object from the queue.
        
        Raises:
            queue.Empty if nothing in the queue.
        """
        if len(self._queue) > 0:
            return self._queue.pop(0)
        else:
            raise queue.Empty
    
    def qsize(self):
        """Returns the current size of the queue."""
        return len(self._queue)
    
    def empty(self):
        """Returns true if the queue is empty."""
        return len(self._queue) == 0


def is_logger(log: object) -> bool:
    """"Returns true if the object is a logger.
    
    Intended to be used by importing module without directly importing logging.
    """
    return isinstance(log, Logger)


def is_log_handler(logger: Logger, handler: object) -> bool:
    """Returns true if the handler is found in the logger.
    
    Args:
        logger: the logger parent of the handler
        handler: the handler to validate
    
    Returns:
        True if the handler is in the logger.

    """
    if not is_logger(logger):
        return False
    found = False
    for h in logger.handlers:
        if h.name == handler.name:
            found = True
            break
    return found


def get_caller_name(depth: int = 2,
                    mod: bool = True,
                    cls: bool =False,
                    mth: bool = False):
    """Returns the name of the calling function.

    For debugging/logging it is often helpful to have a sense of the parentage/flow 
    indicated by *caller_name*.
    
    Args:
        depth: Starting depth of stack inspection.
        mod: Include module name.
        cls: Include class name.
        mth: Include method name.
    
    Returns:
        Name (string) including module[.class][.method]

    """
    stack = inspect.stack()
    start = 0 + depth
    if len(stack) < start + 1:
        return ''
    parent_frame = stack[start][0]
    name = []
    module = inspect.getmodule(parent_frame)
    if module and mod:
        name.append(module.__name__)
    if cls and 'self' in parent_frame.f_locals:
        name.append(parent_frame.f_locals['self'].__class__.__name__)
    if mth:
        codename = parent_frame.f_code.co_name
        if codename != '<module>':
            name.append(codename)
    del parent_frame, stack
    return '.'.join(name)


def get_key_by_value(dictionary: dict, value: any) -> str:
    """Returns the key of the first matching value in the dictionary.
    
    Args:
        dictionary: the dictionary being searched
        value: the value being searched for
    
    Returns:
        The first key with matching value
    
    Raises:
        ValueError if value not found

    """
    for k, v in dictionary:
        if v == value:
            return k
    raise ValueError('Value {} not found in dictionary'.format(value))


def get_wrapping_logger(name: str = None,
                        filename: str = None,
                        file_size: int = 5,
                        max_files: int = 2,
                        debug: bool = False,
                        log_level: int = logging.INFO,
                        **kwargs) -> Logger:
    """Sets up a wrapping logger that writes to console and optionally a file.

    Logging is crucial for debugging embedded systems.  However you don't want
    to fill up an embedded drive with log files, so a *wrapping_logger* is
    useful that:

        * Initializes logging to console, and optionally a CSV formatted file
        * Log file wraps at a given maximum size (default 5 MB)
        * Is easy to configure a logging level (default INFO)
        * Uses UTC/GMT/Zulu timestamps
        * Provides a standardized CSV format
            * ``timestamp,[level],(thread),module.function:line,message``

    Args:
        name: Name of the logger (if None, uses name of calling module).
        filename: (optional) Name of the file/path if writing to a file.
        file_size: Max size of the file in megabytes, before wrapping.
        max_files: The maximum number of files in rotation.
        debug: *backward compatible* enable DEBUG logging
        log_level: A logging level (default INFO)
        kwargs: Optional overrides for RotatingFileHandler
    
    Returns:
        A logger with console stream handler and (optional) file handler.

    """
    FORMAT = ('%(asctime)s.%(msecs)03dZ,[%(levelname)s],(%(threadName)-10s),'
              '%(module)s.%(funcName)s:%(lineno)d,%(message)s')
    log_formatter = logging.Formatter(fmt=FORMAT,
                                      datefmt='%Y-%m-%dT%H:%M:%S')
    log_formatter.converter = gmtime

    if name is None:
        name = get_caller_name()
    logger = logging.getLogger(name)

    if debug or logger.getEffectiveLevel() == logging.DEBUG:
        log_lvl = logging.DEBUG
    else:
        log_lvl = log_level
    logger.setLevel(log_lvl)
    #: Set up log file
    if filename is not None:
        # TODO: validate that logfile is a valid path/filename
        mode = 'a'
        max_bytes = int(file_size * 1024 * 1024)
        backup_count = max_files
        encoding = None
        delay = 0
        for kw in kwargs:
            if kw == 'backupCount':
                backup_count = kwargs[kw]
            elif kw == 'delay':
                delay = kwargs[kw]
            elif kw == 'encoding':
                encoding = kwargs[kw]
            elif kw == 'mode':
                mode = kwargs[kw]
            elif kw == 'maxBytes':
                max_bytes = kwargs[kw]
        file_handler = RotatingFileHandler(
            filename=filename,
            mode=mode,
            maxBytes=max_bytes,
            backupCount=backup_count,
            encoding=encoding,
            delay=delay)
        file_handler.name = name + '_file_handler'
        file_handler.setFormatter(log_formatter)
        file_handler.setLevel(log_lvl)
        if not is_log_handler(logger, file_handler):
            logger.addHandler(file_handler)
    #: Set up console log
    console_handler = logging.StreamHandler()
    console_handler.name = name + '_console_handler'
    console_handler.setFormatter(log_formatter)
    console_handler.setLevel(log_lvl)
    if not is_log_handler(logger, console_handler):
        logger.addHandler(console_handler)

    return logger


def validate_serial_port(target: str, verbose: bool = False) -> Union[bool, tuple]:
    """Validates a given serial port as available on the host.

    When working with different OS and platforms, using a serial port to connect
    to a modem can be simplified by *validate_serial_port*.

    If target port is not found, a list of available ports is returned.
    Labels known FTDI and Prolific serial/USB drivers.

    Args:
        target: Target port name e.g. ``/dev/ttyUSB0``
    
    Returns:
        True or False if detail is False
        (valid: bool, description: str) if detail is True
    """
    found = False
    detail = ''
    ser_ports = [tuple(port) for port in list(list_ports.comports())]
    for port in ser_ports:
        if target == port[0]:
            found = True
            usb_id = str(port[2])
            if 'USB VID:PID=0403:6001' in usb_id:
                driver = 'Serial FTDI FT232 (RS485/RS422/RS232)'
            elif 'USB VID:PID=067B:2303' in usb_id:
                driver = 'Serial Prolific PL2303 (RS232)'
            else:
                driver = 'Serial vendor/device {}'.format(usb_id)
            detail = '{} on {}'.format(driver, port[0])
    if not found and len(ser_ports) > 0:
        for port in ser_ports:
            if len(detail) > 0:
                detail += ','
            detail += " {}".format(port[0])
        detail = 'Available ports:' + detail
    return (found, detail) if verbose else found


if __name__ == '__main__':
    print('This module is not meant to be run directly.')



"""Codec functions for IDP Common Message Format supported by Inmarsat MGS.

Also supported on ORBCOMM IGWS1.
"""

from binascii import b2a_base64
from math import log2, ceil
from struct import pack, unpack
from typing import Tuple, Union
from warnings import WarningMessage, warn
import xml.etree.ElementTree as ET
from xml.dom.minidom import parseString

from idpmodem.constants import FORMAT_HEX, FORMAT_B64, SATELLITE_GENERAL_TRACE

__version__ = '2.0.0'


DATA_TYPES = {
    'bool': 'BooleanField',
    'int_8': 'SignedIntField',
    'uint_8': 'UnsignedIntField',
    'int_16': 'SignedIntField',
    'uint_16': 'UnsignedIntField',
    'int_32': 'SignedIntField',
    'uint_32': 'UnsignedIntField',
    'int_64': 'SignedIntField',
    'uint_64': 'UnsignedIntField',
    'float': 'DataField',
    'double': 'DataField',
    'string': 'StringField',
    'data': 'DataField',
    'enum': 'EnumField',
    'array': 'ArrayField',   # TODO: support for array type
}
XML_NAMESPACE = {
    'xsi': 'http://www.w3.org/2001/XMLSchema-instance',
    'xsd': 'http://www.w3.org/2001/XMLSchema'
}
SIN_RANGE = (16, 255)

for ns in XML_NAMESPACE:
    ET.register_namespace(ns, XML_NAMESPACE[ns])


def optimal_bits(value_range: tuple) -> int:
    """Returns the optimal number of bits for encoding a specified range."""
    if not (isinstance(value_range, tuple) and len(value_range) == 2 and
        value_range[0] <= value_range[1]):
        #: non-compliant
        raise ValueError('value_range must be of form (min, max)')
    total_range = value_range[1] - value_range[0]
    total_range += 1 if value_range[0] == 0 else 0
    optimal_bits = max(1, ceil(log2(value_range[1] - value_range[0])))
    return optimal_bits


def _encode_field_length(length) -> str:
    if length < 128:
        return '0{:07b}'.format(length)
    return '1{:015b}'.format(length)


def _decode_field_length(binstr: str) -> Tuple[int, int]:
    if binstr[0] == '0':
        bit_index = 8
    else:
        bit_index = 16
    length = int(binstr[1:bit_index], 2)
    return (length, bit_index)


def _attribute_equivalence(reference: object,
                           other: object,
                           exclude: "list[str]" = None) -> bool:
    for attr, val in reference.__dict__.items():
        if exclude is not None and attr in exclude:
            continue
        if not hasattr(other, attr) or val != other.__dict__[attr]:
            return False
    return True


def _indent_xml(elem, level=0):
    xmlstr = parseString(ET.tostring(elem)).toprettyxml(indent="  ")
    # i = "\n" + level*"  "
    # j = "\n" + (level-1)*"  "
    # if len(elem):
    #     if not elem.text or not elem.text.strip():
    #         elem.text = i + "  "
    #     if not elem.tail or not elem.tail.strip():
    #         elem.tail = i
    #     for subelem in elem:
    #         _indent_xml(subelem, level+1)
    #     if not elem.tail or not elem.tail.strip():
    #         elem.tail = j
    # else:
    #     if level and (not elem.tail or not elem.tail.strip()):
    #         elem.tail = j
    # return elem
    return xmlstr


class BaseField:
    """The base class for a Field.
    
    Attributes:
        data_type (str): The data type from a supported list.
        name (str): The unique Field name.
        description (str): Optional description.
        optional (bool): Optional indication the field is optional.

    """
    def __init__(self,
                 name: str,
                 data_type: str,
                 description: str = None,
                 optional: bool = False) -> None:
        """Instantiates the base field.
        
        Args:
            name: The field name must be unique within a Message.
            data_type: The data type represented within the field.
            description: (Optional) Description/purpose of the field.
            optional: (Optional) Indicates if the field is mandatory.
            
        """
        if data_type not in DATA_TYPES:
            raise ValueError('Invalid data type {}'.format(data_type))
        if name is None or name.strip() == '':
            raise ValueError('Invalid name must be non-empty')
        self.data_type = data_type
        self.name = name
        self.description = description
        self.optional = optional
    
    def __repr__(self) -> str:
        from pprint import pformat
        return pformat(vars(self), indent=4)
    
    def _base_xml(self) -> ET.Element:
        xsi_type = DATA_TYPES[self.data_type]
        xmlfield = ET.Element('Field', attrib={
            '{http://www.w3.org/2001/XMLSchema-instance}type': xsi_type
        })
        name = ET.SubElement(xmlfield, 'Name')
        name.text = self.name
        if self.description:
            description = ET.SubElement(xmlfield, 'Description')
            description.text = str(self.description)
        if self.optional:
            optional = ET.SubElement(xmlfield, 'Optional')
            optional.text = 'true'
        return xmlfield


class Message:
    """The Payload structure for Message Definition Files uploaded to a Mailbox.
    
    Attributes:
        name (str): The message name
        sin (int): The Service Identification Number
        min (int): The Message Identification Number
        fields (list): A list of Fields
        description (str): Optional description
        is_forward (bool): Indicates if the message is mobile-terminated

    """

    def __init__(self,
                 name: str,
                 sin: int,
                 min: int,
                 description: str = None,
                 is_forward: bool = False,
                 fields: "list[BaseField]" = None):
        """Instantiates a Message.
        
        Args:
            name: The message name should be unique within the xMessages list.
            sin: The Service Identification Number (16..255)
            min: The Message Identification Number (0..255)
            description: (Optional) Description/purpose of the Message.
            is_forward: Indicates if the message is intended to be
                Mobile-Terminated.
            fields: Optional definition of fields during instantiation.

        """
        if not isinstance(sin, int) or sin not in range(16, 256):
            raise ValueError('Invalid SIN {} must be in range 16..255'.format(
                             sin))
        if not isinstance(min, int) or min not in range (0, 256):
            raise ValueError('Invalid MIN {} must be in range 0..255'.format(
                             min))
        self.name = name
        self.description = description
        self.is_forward = is_forward
        self.sin = sin
        self.min = min
        self.fields = fields or Fields()

    @property
    def fields(self) -> list:
        return self._fields
    
    @fields.setter
    def fields(self, fields: "list[BaseField]"):
        if not all(isinstance(field, BaseField) for field in fields):
            raise ValueError('Invalid field found in list')
        self._fields = fields

    @property
    def ota_size(self) -> int:
        ota_bits = 2 * 8
        for field in self.fields:
            ota_bits += field.bits + (1 if field.optional else 0)
        return ceil(ota_bits / 8)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Message):
            return NotImplemented
        return _attribute_equivalence(self, other)

    def decode(self, databytes: bytes) -> None:
        """Parses and stores field values from raw data (received over-the-air).
        
        Args:
            databytes: A bytes array (typically from the forward message)
        """
        binary_str = ''.join(format(int(b), '08b') for b in databytes)
        bit_offset = 16   #: Begin after SIN/MIN bytes
        for field in self.fields:
            if field.optional:
                present = binary_str[bit_offset] == '1'
                bit_offset += 1
                if not present:
                    continue
            bit_offset += field.decode(binary_str[bit_offset:])

    def derive(self, databytes: bytes) -> None:
        """Derives field values from raw data bytes (received over-the-air).
        
        Deprecated/replaced by `decode`.

        Args:
            databytes: A bytes array (typically from the forward message)

        """
        self.decode(databytes)

    def encode(self,
               data_format: int = FORMAT_B64,
               exclude: list = None) -> dict:
        """Encodes using the specified data format (base64 or hex).

        Args:
            data_format (int): 2=ASCII-Hex, 3=base64
            exclude (list[str]): A list of optional field names to exclude
        
        Returns:
            Dictionary with sin, min, data_format and data to pass into AT%MGRT
                or atcommand function `message_mo_send`
        """
        if data_format not in [FORMAT_B64, FORMAT_HEX]:
            raise ValueError('data_format {} unsupported'.format(data_format))
        #:AT%MGRT uses '{}.{},{},{}'.format(sin, min, data_format, data)
        bin_str = ''
        for field in self.fields:
            if field.optional:
                if exclude is not None and field.name in exclude:
                    present = False
                elif hasattr(field, 'value'):
                    present = field.value is not None
                elif hasattr(field, 'elements'):
                    present = field.elements is not None
                else:
                    raise ValueError('Unknown value of optional')
                bin_str += '1' if present else '0'
                if not present:
                    continue
            bin_str += field.encode()
        for _ in range(0, 8 - len(bin_str) % 8):   #:pad to next byte
            bin_str += '0'
        _format = '0{}X'.format(int(len(bin_str) / 8 * 2))   #:hex bytes 2 chars
        hex_str = format(int(bin_str, 2), _format)
        if (self.is_forward and len(hex_str) / 2 > 9998 or
            not self.is_forward and len(hex_str) / 2 > 6398):
            raise ValueError('{} bytes exceeds maximum size for Payload'.format(
                             len(hex_str) / 2))
        if data_format == FORMAT_HEX:
            data = hex_str
        else:
            data = b2a_base64(bytearray.fromhex(hex_str)).strip().decode()
        return {
            'sin': self.sin,
            'min': self.min,
            'data_format': data_format,
            'data': data
        }

    def xml(self, indent: bool = False) -> ET.Element:
        """Returns the XML definition for a Message Definition File."""
        xmessage = ET.Element('Message')
        name = ET.SubElement(xmessage, 'Name')
        name.text = self.name
        min = ET.SubElement(xmessage, 'MIN')
        min.text = str(self.min)
        fields = ET.SubElement(xmessage, 'Fields')
        for field in self.fields:
            fields.append(field.xml())
        return xmessage if not indent else _indent_xml(xmessage)


class Service:
    """A data structure holding a set of related Forward and Return Messages.
    
    Attributes:
        name (str): The service name
        sin (int): Service Identification Number or codec service id (16..255)
        description (str): A description of the service (unsupported)
        messages_forward (list): A list of mobile-terminated Message definitions
        messages_return (list): A list of mobile-originated Message definitions

    """
    def __init__(self,
                 name: str,
                 sin: int,
                 description: str = None,
                 messages_forward: "list[Message]" = None,
                 messages_return: "list[Message]" = None) -> None:
        """Instantiates a Service made up of Messages.
        
        Args:
            name: The service name should be unique within a MessageDefinitions
            sin: The Service Identification Number (16..255)
            description: (Optional)
        """
        if not isinstance(name, str) or name == '':
            raise ValueError('Invalid service name {}'.format(name))
        if sin not in range(16, 256):
            raise ValueError('Invalid SIN must be 16..255')
        self.name = name
        self.sin = sin
        if description is not None:
            warn('Service Description not currently supported')
        self.description = None
        self.messages_forward = messages_forward or Messages(self.sin,
                                                             is_forward=True)
        self.messages_return = messages_return or Messages(self.sin,
                                                           is_forward=False)
    
    def xml(self, indent: bool = False) -> ET.Element:
        """Returns the XML structure of the Service for a MDF."""
        if len(self.messages_forward) == 0 and len(self.messages_return) == 0:
            raise ValueError('No messages defined for service {}'.format(
                             self.sin))
        xservice = ET.Element('Service')
        name = ET.SubElement(xservice, 'Name')
        name.text = str(self.name)
        sin = ET.SubElement(xservice, 'SIN')
        sin.text = str(self.sin)
        if self.description:
            desc = ET.SubElement(xservice, 'Description')
            desc.text = str(self.description)
        if len(self.messages_forward) > 0:
            forward_messages = ET.SubElement(xservice, 'ForwardMessages')
            for m in self.messages_forward:
                forward_messages.append(m.xml())
        if len(self.messages_return) > 0:
            return_messages = ET.SubElement(xservice, 'ReturnMessages')
            for m in self.messages_return:
                return_messages.append(m.xml())
        return xservice if not indent else _indent_xml(xservice)


class ObjectList(list):
    """Base class for a specific object type list.
    
    Used for Fields, Messages, Services.

    Attributes:
        list_type: The object type the list is comprised of.

    """
    SUPPORTED_TYPES = [
        BaseField,
        # Field,
        Message,
        Service,
    ]

    def __init__(self, list_type: object):
        if list_type not in self.SUPPORTED_TYPES:
            raise ValueError('Unsupported object type {}'.format(list_type))
        super().__init__()
        self.list_type = list_type

    def add(self, obj: object) -> bool:
        """Add an object to the end of the list.

        Args:
            obj (object): A valid object according to the list_type
        
        Raises:
            ValueError if there is a duplicate or invalid name,
                invalid value_range or unsupported data_type
        """
        if not isinstance(obj, self.list_type):
            raise ValueError('Invalid {} definition'.format(self.list_type))
        for o in self:
            if o.name == obj.name:
                raise ValueError('Duplicate {} name {} found'.format(
                                 self.list_type, obj.name))
        self.append(obj)
        return True

    def __getitem__(self, n: Union[str, int]) -> object:
        """Retrieves an object by name or index.
        
        Args:
            n: The object name or list index
        
        Returns:
            object

        """
        if isinstance(n, str):
            for o in self:
                if o.name == n:
                    return o
            raise ValueError('{} name {} not found'.format(self.list_type, n))
        return super().__getitem__(n)

    def __setitem__(self, n: Union[str, int], value):
        if isinstance(n, str):
            for o in self:
                if o.name == n:
                    o.value = value
                    break
        else:
            super().__setitem__(n, value)

    def delete(self, name: str) -> bool:
        """Delete an object from the list by name.
        
        Args:
            name: The name of the object.

        Returns:
            boolean: success
        """
        for o in self:
            if o.name == name:
                self.remove(o)
                return True
        return False


class Fields(ObjectList):
    """The list of Fields defining a Message or ArrayElement."""
    def __init__(self, fields: "list[BaseField]" = None):
        super().__init__(list_type=BaseField)
        if fields is not None:
            for field in fields:
                self.add(field)
    
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Fields):
            return NotImplemented
        if len(self) != len(other):
            return False
        for field in self:
            if field != other[field.name]:
                return False
        return True


class Messages(ObjectList):
    """The list of Messages (Forward or Return) within a Service."""
    def __init__(self, sin: int, is_forward: bool):
        super().__init__(list_type=Message)
        self.sin = sin
        self.is_forward = is_forward
    
    def add(self, message: Message) -> bool:
        """Add a message to the list if it matches the parent SIN.

        Overrides the base class add method.

        Args:
            message (object): A valid Message
        
        Raises:
            ValueError if there is a duplicate or invalid name,
                invalid value_range or unsupported data_type

        """
        if not isinstance(message, Message):
            raise ValueError('Invalid message definition')
        if message.sin != self.sin:
            raise ValueError('Message SIN {} does not match service {}'.format(
                             message.sin, self.sin))
        for m in self:
            if m.name == message.name:
                raise ValueError('Duplicate message name {} found'.format(
                                 message.name))
            if m.min == message.min:
                raise ValueError('Duplicate message MIN {} found'.format(
                                 message.min))
        self.append(message)
        return True


class Services(ObjectList):
    """The list of Service(s) within a MessageDefinitions."""
    def __init__(self, services: "list[Service]" = None):
        super().__init__(list_type=Service)
        if services is not None:
            for service in services:
                if not isinstance(service, Service):
                    raise ValueError('Invalid Service {}'.format(service))
                self.add(service)
    
    def add(self, service: Service) -> None:
        """Adds a Service to the list of Services."""
        if not isinstance(service, Service):
            raise ValueError('{} is not a valid Service'.format(service))
        if service.name in self:
            raise ValueError('Duplicate Service {}'.format(service.name))
        for existing_service in self:
            if existing_service.sin == service.sin:
                raise ValueError('Duplicate SIN {}'.format(service.sin))
        self.append(service)


class BooleanField(BaseField):
    """A Boolean field."""
    def __init__(self,
                 name: str,
                 description: str = None,
                 optional: bool = False,
                 default: bool = False,
                 value: bool = None) -> None:
        super().__init__(name=name,
                         data_type='bool',
                         description=description,
                         optional=optional)
        """Instantiates a BooleanField.
        
        Args:
            name: The field name must be unique within a Message.
            description: An optional description/purpose for the field.
            optional: Indicates if the field is optional in the Message.
            default: A default value for the boolean.
            value: Optional value to set during initialization.

        """
        self.default = default
        self.value = value if value is not None else default
    
    @property
    def default(self):
        return self._default

    @default.setter
    def default(self, v: bool):
        if v is not None and not isinstance(v, bool):
            raise ValueError('Invalid boolean value {}'.format(v))
        self._default = v

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, v: bool):
        if v is not None and not isinstance(v, bool):
            raise ValueError('Invalid boolean value {}'.format(v))
        self._value = v

    @property
    def bits(self):
        return 1
    
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BooleanField):
            return NotImplemented
        return _attribute_equivalence(self, other)

    def encode(self) -> str:
        """Returns the binary string of the field value."""
        if self.value is None and not self.optional:
            raise ValueError('No value assigned to field')
        return '1' if self.value else '0'

    def decode(self, binary_str: str) -> int:
        """Populates the field value from binary and returns the next offset.
        
        Args:
            binary_str (str): The binary string to decode
        
        Returns:
            The bit offset after parsing
        """
        self.value = True if binary_str[0] == '1' else False
        return 1

    def xml(self) -> ET.Element:
        xmlfield = self._base_xml()
        if self.default:
            default = ET.SubElement(xmlfield, 'Default')
            default.text = 'true'
        return xmlfield


class EnumField(BaseField):
    """An enumerated field sends an index over-the-air representing a string."""
    def __init__(self,
                 name: str,
                 items: "list[str]",
                 size: int,
                 description: str = None,
                 optional: bool = False,
                 default: int = None,
                 value: int = None) -> None:
        """Instantiates a EnumField.
        
        Args:
            name: The field name must be unique within a Message.
            items: A list of strings (indexed from 0).
            size: The number of *bits* used to encode the index over-the-air.
            description: An optional description/purpose for the field.
            optional: Indicates if the field is optional in the Message.
            default: A default value for the enum.
            value: Optional value to set during initialization.

        """
        super().__init__(name=name,
                         data_type='enum',
                         description=description,
                         optional=optional)
        if items is None or not all(isinstance(item, str) for item in items):
            raise ValueError('Items must all be strings')
        if not isinstance(size, int) or size < 1:
            raise ValueError('Size must be integer greater than 0 bits')
        self.items = items
        self.size = size
        self.default = default
        self.value = value if value is not None else self.default
    
    def _validate_enum(self, v: Union[int, str]) -> Union[int, None]:
        if v is not None:
            if isinstance(v, str):
                if v not in self.items:
                    raise ValueError('Invalid value {}'.format(v))
                for index, item in enumerate(self.items):
                    if item == v:
                        return index
            elif isinstance(v, int):
                if v < 0 or v >= len(self.items):
                    raise ValueError('Invalid enum index {}'.format(v))
            else:
                raise ValueError('Invalid value {}'.format(v))
        return v

    @property
    def items(self):
        return self._items
    
    @items.setter
    def items(self, l: list):
        if not isinstance(l, list) or not all(isinstance(x, str) for x in l):
            raise ValueError('Items must be a list of strings')
        self._items = l

    @property
    def default(self):
        if self._default is None:
            return None
        return self.items[self._default]
    
    @default.setter
    def default(self, v: Union[int, str]):
        self._default = self._validate_enum(v)

    @property
    def value(self):
        if self._value is None:
            return None
        return self.items[self._value]
    
    @value.setter
    def value(self, v: Union[int, str]):
        self._value = self._validate_enum(v)

    @property
    def size(self):
        return self._size
    
    @size.setter
    def size(self, v: int):
        if not isinstance(v, int) or v < 1:
            raise ValueError('Size must be integer greater than zero')
        minimum_bits = optimal_bits((0, len(self.items)))
        if v < minimum_bits:
            raise ValueError('Size must be at least {} to support item count'
                             .format(minimum_bits))
        self._size = v

    @property
    def bits(self):
        return self.size
    
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, EnumField):
            return NotImplemented
        return _attribute_equivalence(self, other)

    def encode(self) -> str:
        """Returns the binary string of the field value."""
        if self.value is None:
            raise ValueError('No value configured in EnumField {}'.format(
                             self.name))
        _format = '0{}b'.format(self.bits)
        binstr = format(self.items.index(self.value), _format)
        return binstr

    def decode(self, binary_str: str) -> int:
        """Populates the field value from binary and returns the next offset.
        
        Args:
            binary_str (str): The binary string to decode
        
        Returns:
            The bit offset after parsing
        """
        self.value = binary_str[:self.bits]
        return self.bits

    def xml(self) -> ET.Element:
        xmlfield = self._base_xml()
        size = ET.SubElement(xmlfield, 'Size')
        size.text = str(self.size)
        items = ET.SubElement(xmlfield, 'Items')
        for string in self.items:
            item = ET.SubElement(items, 'string')
            item.text = str(string)
        if self.default:
            default = ET.SubElement(xmlfield, 'Default')
            default.text = str(self.default)
        return xmlfield


class UnsignedIntField(BaseField):
    """An unsigned integer value using a defined number of bits over-the-air."""
    def __init__(self,
                 name: str,
                 size: int,
                 data_type: str = 'uint_16',
                 description: str = None,
                 optional: bool = False,
                 default: int = None,
                 value: int = None) -> None:
        """Instantiates a UnsignedIntField.
        
        Args:
            name: The field name must be unique within a Message.
            size: The number of *bits* used to encode the integer over-the-air
                (maximum 32).
            data_type: The integer type represented (for decoding).
            description: An optional description/purpose for the string.
            optional: Indicates if the string is optional in the Message.
            default: A default value for the string.
            value: Optional value to set during initialization.

        """
        if data_type not in ['uint_8', 'uint_16', 'uint_32']:
            raise ValueError('Invalid unsignedint type {}'.format(data_type))
        super().__init__(name=name,
                         data_type=data_type,
                         description=description,
                         optional=optional)
        self.size = size
        self.default = default
        self.value = value if value is not None else default
    
    @property
    def size(self):
        return self._size

    @size.setter
    def size(self, value: int):
        if not isinstance(value, int) or value < 1:
            raise ValueError('Size must be integer greater than 0 bits')
        data_type_size = int(self.data_type.split('_')[1])
        if value > data_type_size:
            warn('Size {} larger than required by {}'.format(
                value, self.data_type))
        self._size = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, v: int):
        clip = False
        if v is not None:
            if not isinstance(v, int) or v < 0:
                raise ValueError('Unsignedint must be non-negative integer')
            if v > 2**self.size - 1:
                self._value = 2**self.size - 1
                warn('Clipping unsignedint at max value {}'.format(self._value))
                clip = True
        if not clip:
            self._value = v
    
    @property
    def default(self):
        return self._default
    
    @default.setter
    def default(self, v: int):
        if v is not None:
            if v > 2**self.size - 1 or v < 0:
                raise ValueError('Invalid unsignedint default {}'.format(v))
        self._default = v
    
    @property
    def bits(self):
        return self.size
    
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, UnsignedIntField):
            return NotImplemented
        return _attribute_equivalence(self, other)

    def encode(self) -> str:
        """Returns the binary string of the field value."""
        if self.value is None:
            raise ValueError('No value defined in UnsignedIntField {}'.format(
                             self.name))
        _format = '0{}b'.format(self.bits)
        return format(self.value, _format)

    def decode(self, binary_str: str) -> int:
        """Populates the field value from binary and returns the next offset.
        
        Args:
            binary_str (str): The binary string to decode
        
        Returns:
            The bit offset after parsing
        """
        self.value = int(binary_str[:self.bits], 2)
        return self.bits

    def xml(self) -> ET.Element:
        xmlfield = self._base_xml()
        size = ET.SubElement(xmlfield, 'Size')
        size.text = str(self.size)
        if self.default:
            default = ET.SubElement(xmlfield, 'Default')
            default.text = str(self.default)
        return xmlfield


class SignedIntField(BaseField):
    """A signed integer value using a defined number of bits over-the-air."""
    def __init__(self,
                 name: str,
                 size: int,
                 data_type: str = 'int_16',
                 description: str = None,
                 optional: bool = False,
                 default: int = None,
                 value: int = None) -> None:
        """Instantiates a SignedIntField.
        
        Args:
            name: The field name must be unique within a Message.
            size: The number of *bits* used to encode the integer over-the-air
                (maximum 32).
            data_type: The integer type represented (for decoding).
            description: An optional description/purpose for the string.
            optional: Indicates if the string is optional in the Message.
            default: A default value for the string.
            value: Optional value to set during initialization.

        """
        if data_type not in ['int_8', 'int_16', 'int_32']:
            raise ValueError('Invalid unsignedint type {}'.format(data_type))
        super().__init__(name=name,
                         data_type=data_type,
                         description=description,
                         optional=optional)
        self.size = size
        self.default = default
        self.value = value if value is not None else default
    
    @property
    def size(self):
        return self._size

    @size.setter
    def size(self, value: int):
        if not isinstance(value, int) or value < 1:
            raise ValueError('Size must be integer greater than 0 bits')
        self._size = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, v: int):
        clip = False
        if v is not None:
            if not isinstance(v, int):
                raise ValueError('Unsignedint must be non-negative integer')
            if v > (2**self.size / 2) - 1:
                self._value = int(2**self.size / 2) - 1
                warn('Clipping signedint at max value {}'.format(self._value))
                clip = True
            if v < -(2**self.size / 2):
                self._value = -1 * int(2**self.size / 2)
                warn('Clipping signedint at min value {}'.format(self._value))
                clip = True
        if not clip:
            self._value = v
    
    @property
    def default(self):
        return self._default
    
    @default.setter
    def default(self, v: int):
        if v is not None:
            if not isinstance(v, int):
                raise ValueError('Invalid signed integer {}'.format(v))
            if v > (2**self.size / 2) - 1 or v < -(2**self.size / 2):
                raise ValueError('Invalid default {}'.format(v))
        self._default = v
    
    @property
    def bits(self):
        return self.size
    
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SignedIntField):
            return NotImplemented
        return _attribute_equivalence(self, other)

    def encode(self) -> str:
        """Returns the binary string of the field value."""
        if self.value is None:
            raise ValueError('No value defined in UnsignedIntField {}'.format(
                             self.name))
        _format = '0{}b'.format(self.bits)
        if self.value < 0:
            invertedbin = format(self.value * -1, _format)
            twocomplementbin = ''
            i = 0
            while len(twocomplementbin) < len(invertedbin):
                twocomplementbin += '1' if invertedbin[i] == '0' else '0'
                i += 1
            binstr = format(int(twocomplementbin, 2) + 1, _format)
        else:
            binstr = format(self.value, _format)
        return binstr

    def decode(self, binary_str: str) -> int:
        """Populates the field value from binary and returns the next offset.
        
        Args:
            binary_str (str): The binary string to decode
        
        Returns:
            The bit offset after parsing
        """
        value = int(binary_str[:self.bits], 2)
        if (value & (1 << (self.bits - 1))) != 0:   #:sign bit set e.g. 8bit: 128-255
            value = value - (1 << self.bits)        #:compute negative value
        self.value = value
        return self.bits

    def xml(self) -> ET.Element:
        xmlfield = self._base_xml()
        size = ET.SubElement(xmlfield, 'Size')
        size.text = str(self.size)
        if self.default:
            default = ET.SubElement(xmlfield, 'Default')
            default.text = str(self.default)
        return xmlfield


class StringField(BaseField):
    """A character string sent over-the-air."""
    def __init__(self,
                 name: str,
                 size: int,
                 description: str = None,
                 optional: bool = False,
                 fixed: bool = False,
                 default: str = None,
                 value: str = None) -> None:
        """Instantiates a StringField.
        
        Args:
            name: The field name must be unique within a Message.
            size: The maximum number of characters in the string.
            description: An optional description/purpose for the string.
            optional: Indicates if the string is optional in the Message.
            fixed: Indicates if the string is always fixed length `size`.
            default: A default value for the string.
            value: Optional value to set during initialization.

        """
        super().__init__(name=name,
                         data_type='string',
                         description=description,
                         optional=optional)
        self.size = size
        self.fixed = fixed
        self.default = default
        self.value = value if value is not None else default
    
    def _validate_string(self, s: str):
        if s is not None:
            if not isinstance(s, str):
                raise ValueError('Invalid string {}'.format(s))
            if len(s) > self.size:
                warn('Clipping string at max {} characters'.format(self.size))
                return s[:self.size]
        return s
                
    @property
    def size(self):
        return self._size
    
    @size.setter
    def size(self, value: int):
        if not isinstance(value, int) or value < 1:
            raise ValueError('Size must be integer greater than 0 characters')
        self._size = value
    
    @property
    def default(self) -> str:
        return self._default
    
    @default.setter
    def default(self, v: str):
        self._default = self._validate_string(v)

    @property
    def value(self) -> str:
        return self._value
    
    @value.setter
    def value(self, v: str):
        self._value = self._validate_string(v)

    @property
    def bits(self):
        if self.fixed or self.value is None:
            return self.size * 8
        return len(self.value) * 8
    
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, StringField):
            return NotImplemented
        return _attribute_equivalence(self, other)

    def encode(self) -> str:
        """Returns the binary string of the field value."""
        if self.value is None and not self.optional:
            raise ValueError('No value defined for StringField {}'.format(
                             self.name))
        binstr = ''.join(format(ord(c), '08b') for c in self.value)
        if self.fixed:
            binstr += ''.join('0' for bit in range(len(binstr), self.bits))
        else:
            binstr = _encode_field_length(len(self.value)) + binstr
        return binstr

    def decode(self, binary_str: str) -> int:
        """Populates the field value from binary and returns the next offset.
        
        Args:
            binary_str (str): The binary string to decode
        
        Returns:
            The bit offset after parsing
        """
        if self.fixed:
            length = self.size
            bit_index = 0
        else:
            (length, bit_index) = _decode_field_length(binary_str)
        n = int(binary_str[bit_index:bit_index + length * 8], 2)
        char_bytes = n.to_bytes((n.bit_length() + 7) // 8, 'big')
        for i, byte in enumerate(char_bytes):
            if byte == 0:
                warn('Truncating after 0 byte in string')
                char_bytes = char_bytes[:i]
                break
        self.value = char_bytes.decode('utf-8', 'surrogatepass') or '\0'
        return bit_index + length * 8

    def xml(self) -> ET.Element:
        """Returns the message definition XML representation of the field."""
        xmlfield = self._base_xml()
        size = ET.SubElement(xmlfield, 'Size')
        size.text = str(self.size)
        if self.fixed:
            fixed = ET.SubElement(xmlfield, 'Fixed')
            fixed.text = 'true'
        if self.default:
            default = ET.SubElement(xmlfield, 'Default')
            default.text = str(self.default)
        return xmlfield


class DataField(BaseField):
    """A data field of raw bytes sent over-the-air.
    
    Can also be used to hold floating point, double-precision or large integers.

    """
    supported_data_types = ['data', 'float', 'double']
    def __init__(self,
                 name: str,
                 size: int,
                 data_type: str = 'data',
                 description: str = None,
                 optional: bool = False,
                 fixed: bool = False,
                 default: bytes = None,
                 value: bytes = None) -> None:
        """Instantiates a EnumField.
        
        Args:
            name: The field name must be unique within a Message.
            size: The maximum number of bytes to send over-the-air.
            data_type: The data type represented within the bytes.
            description: An optional description/purpose for the field.
            optional: Indicates if the field is optional in the Message.
            fixed: Indicates if the data bytes are a fixed `size`.
            default: A default value for the enum.
            value: Optional value to set during initialization.

        """
        if data_type is None or data_type not in self.supported_data_types:
            raise ValueError('Invalid data type {}'.format(data_type))
        super().__init__(name=name,
                         data_type=data_type,
                         description=description,
                         optional=optional)
        self.fixed = fixed
        self.size = size
        self.default = default
        self.value = value if value is not None else default
    
    @property
    def size(self):
        return self._size
    
    @size.setter
    def size(self, value: int):
        if not isinstance(value, int) or value < 1:
            raise ValueError('Size must be integer greater than 0 bytes')
        if self.data_type == 'float':
            if value != 4:
                warn('Adjusting float size to 4 bytes fixed')
            self._size = 4
            self.fixed = True
        elif self.data_type == 'double':
            if value != 8:
                warn('Adjusting float size to 8 bytes fixed')
            self._size = 4
            self.fixed = True
        else:
            self._size = value
    
    @property
    def value(self):
        if self.data_type == 'float':
            return unpack('!f', self._value)
        if self.data_type == 'double':
            return unpack('!d', self._value)
        return self._value

    @value.setter
    def value(self, v: Union[bytes, float]):
        if v is not None:
            if self.data_type in ['float', 'double']:
                if not isinstance(v, float):
                    raise ValueError('Invalid {} value {}'.format(
                                     self.data_type, v))
                _format = '!f' if self.data_type == 'float' else '!d'
                v = pack(_format, v)
            elif not isinstance(v, bytes):
                raise ValueError('Invalid bytes {}'.format(v))
        self._value = v

    @property
    def bits(self):
        if self.fixed:
            return self.size * 8
        elif self.value is None:
            return 0
        return len(self.value) * 8
    
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DataField):
            return NotImplemented
        return _attribute_equivalence(self, other)

    def encode(self) -> str:
        """Returns the binary string of the field value."""
        if self.value is None and not self.optional:
            raise ValueError('No value defined for DataField {}'.format(
                             self.name))
        binstr = ''
        binstr = ''.join(format(b, '08b') for b in self._value)
        if self.fixed:   #:pad to fixed length
            binstr += ''.join('0' for bit in range(len(binstr), self.bits))
        else:
            binstr = _encode_field_length(len(self._value)) + binstr
        return binstr

    def decode(self, binary_str: str) -> int:
        """Populates the field value from binary and returns the next offset.
        
        Args:
            binary_str (str): The binary string to decode
        
        Returns:
            The bit offset after parsing
        """
        if self.fixed:
            binary = binary_str[:self.bits]
            bits = self.bits
        else:
            (length, bit_index) = _decode_field_length(binary_str)
            binary = binary_str[bit_index:length * 8 + bit_index]
            bits = len(binary)
        self._value = int(binary, 2).to_bytes(int(bits / 8), 'big')
        return self.bits

    def xml(self) -> ET.Element:
        xmlfield = self._base_xml()
        size = ET.SubElement(xmlfield, 'Size')
        size.text = str(self.size)
        if self.default:
            default = ET.SubElement(xmlfield, 'Default')
            default.text = str(self.default)
        return xmlfield


class ArrayField(BaseField):
    """An Array Field provides a list where each element is a set of Fields.
    
    Attributes:
        name (str): The name of the field instance.
        size (int): The maximum number of elements allowed.
        fields (Fields): A list of Field types comprising each ArrayElement
        description (str): An optional description of the array/use.
        optional (bool): Indicates if the array is optional in the Message
        fixed (bool): Indicates if the array is always the fixed `size`
        elements (list): The enumerated list of ArrayElements

    """
    def __init__(self,
                 name: str,
                 size: int,
                 fields: Fields,
                 description: str = None,
                 optional: bool = False,
                 fixed: bool = False,
                 elements: "list[Fields]" = None) -> None:
        """Initializes an ArrayField instance.
        
        Args:
            name: The unique field name within the Message.
            size: The maximum number of elements allowed.
            fields: The list of Field types comprising each element.
            description: An optional description/purpose of the array.
            optional: Indicates if the array is optional in the Message.
            fixed: Indicates if the array is always the fixed `size`.
            elements: Option to populate elements of Fields during instantiation.

        """
        super().__init__(name=name,
                         data_type='array',
                         description=description,
                         optional=optional)
        self.size = size
        self.fixed = fixed
        self.fields = fields
        self.elements = elements or []
    
    @property
    def size(self):
        return self._size
    
    @size.setter
    def size(self, value: int):
        if not isinstance(value, int) or value < 1:
            raise ValueError('Size must be integer greater than 0 fields')
        self._size = value
    
    @property
    def fields(self):
        return self._fields

    @fields.setter
    def fields(self, fields: Fields):
        if not isinstance(fields, Fields):
            raise ValueError('Invalid Fields definition for ArrayField')
        self._fields = fields

    @property
    def elements(self):
        return self._elements
    
    @elements.setter
    def elements(self, elements: list):
        if elements is None or not hasattr(self, '_elements'):
            self._elements = []
        if not isinstance(elements, list):
            raise ValueError('Elements must be a list of grouped Fields')
        for fields in elements:
            for index, field in enumerate(fields):
                if (field.name != self.fields[index].name):
                    raise ValueError('fields[{}].name expected {} got {}'
                                     .format(index,
                                             self.fields[index].name,
                                             field.name))
                if (field.data_type != self.fields[index].data_type):
                    raise ValueError('fields[{}].data_type expected {} got {}'
                                     .format(index,
                                             self.fields[index].data_type,
                                             field.data_type))
                #TODO: validate non-optional fields have value/elements
                if field.value is None and not field.optional:
                    raise ValueError('fields[{}].value missing'.format(index))
                try:
                    self._elements[index] = fields
                except IndexError:
                    self._elements.append(fields)

    @property
    def bits(self):
        bits = 0
        for field in self.fields:
            bits += field.bits
        return bits
    
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ArrayField):
            return NotImplemented
        return _attribute_equivalence(self, other)

    def _valid_element(self, element: Fields) -> bool:
        for i, field in enumerate(self.fields):
            e = element[i]
            if e.name != field.name:
                raise ValueError('element field name {} does not match {}'
                    .format(e.name, field.name))
            if e.data_type != field.data_type:
                raise ValueError('element field data_type {} does not match {}'
                    .format(e.data_type, field.data_type))
            if e.optional != field.optional:
                raise ValueError('element optional {} does not match {}'
                    .format(e.optional, field.optional))
            if hasattr(field, 'fixed') and e.fixed != field.fixed:
                raise ValueError('element fixed {} does not match {}'
                    .format(e.fixed, field.fixed))
            if hasattr(field, 'size') and e.size != field.size:
                raise ValueError('element size {} does not match {}'
                    .format(e.size, field.size))
        return True

    def append(self, element: Fields):
        """Adds the array element to the list of elements."""
        if not isinstance(element, Fields):
            raise ValueError('Invalid element definition must be Fields')
        if self._valid_element(element):
            for i, field in enumerate(element):
                if (hasattr(field, 'description') and
                    field.description != self.fields[i].description):
                    field.description = self.fields[i].description
                if hasattr(field, 'value') and field.value is None:
                    field.value = self.fields[i].default
            self._elements.append(element)

    def new_element(self) -> Fields:
        """Returns an empty element at the end of the elements list."""
        new_index = len(self._elements)
        self._elements.append(Fields(self.fields))
        return self.elements[new_index]

    def encode(self) -> str:
        """Returns the binary string of the field value."""
        if len(self.elements) == 0:
            raise ValueError('No elements to encode')
        binstr = ''
        for element in self.elements:
            for field in element:
                binstr += field.encode()
        if not self.fixed:
            binstr = _encode_field_length(len(self.elements)) + binstr
        return binstr

    def decode(self, binary_str: str) -> int:
        """Populates the field value from binary and returns the next offset.
        
        Args:
            binary_str (str): The binary string to decode
        
        Returns:
            The bit offset after parsing
        """
        if self.fixed:
            length = self.size
            bit_index = 0
        else:
            (length, bit_index) = _decode_field_length(binary_str)
        for index in range(0, length):
            fields = Fields(self.fields)
            for field in fields:
                if field.optional:
                    if binary_str[bit_index] == '0':
                        bit_index += 1
                        continue
                    bit_index += 1
                bit_index += field.decode(binary_str[bit_index:])
            try:
                self._elements[index] = fields
            except IndexError:
                self._elements.append(fields)
        return bit_index

    def xml(self) -> ET.Element:
        xmlfield = self._base_xml()
        size = ET.SubElement(xmlfield, 'Size')
        size.text = str(self.size)
        if self.fixed:
            default = ET.SubElement(xmlfield, 'Fixed')
            default.text = 'true'
        fields = ET.SubElement(xmlfield, 'Fields')
        for field in self.fields:
            fields.append(field.xml())
        return xmlfield


class MessageDefinitions:
    """A set of Message Definitions grouped into Services.

    Attributes:
        services: The list of Services with Messages defined.
    
    """
    def __init__(self, services: Services = None):
        if services is not None:
            if not isinstance(services, Services):
                raise ValueError('Invalid Services')
        self.services = services or Services()
    
    def xml(self, indent: bool = False) -> ET.Element:
        xmsgdef = ET.Element('MessageDefinition',
                             attrib={'xmlns:xsd': XML_NAMESPACE['xsd']})
        services = ET.SubElement(xmsgdef, 'Services')
        for service in self.services:
            services.append(service.xml())
        return xmsgdef if not indent else _indent_xml(xmsgdef)
    
    def mdf_export(self, filename: str, pretty: bool = False):
        tree = ET.ElementTree(self.xml())
        root = tree.getroot()
        if pretty:
            from xml.dom.minidom import parseString
            xmlstr = parseString(ET.tostring(root)).toprettyxml(indent="  ")
            with open(filename, 'w') as f:
                f.write(xmlstr)
        else:
            with open(filename, 'wb') as f:
                tree.write(f, encoding='utf-8', xml_declaration=True)



"""Message classes for abstracting messages presented by/to the AT commands.
"""

from base64 import b64decode, b64encode
import binascii
from logging import Logger
from string import printable
from time import time

from idpmodem.utils import get_wrapping_logger
from idpmodem.constants import FORMAT_B64, FORMAT_HEX, FORMAT_TEXT, PRIORITY_LOW, PRIORITY_MT, RX_RETRIEVED


def _is_hex_string(s: str) -> bool:
    """Returns True if the string consists exclusively of hexadecimal chars."""
    hex_chars = '0123456789abcdefABCDEF'
    return all(c in hex_chars for c in s)


def _bytearray_to_str(arr: bytearray) -> str:
    """Converts a bytearray to a readable text string."""
    s = ''
    for b in bytearray(arr):
        if chr(b) in printable:
            s += chr(b)
        else:
            s += '{0:#04x}'.format(b).replace('0x', '\\')
    return s


def _bytearray_to_hex_str(arr: bytearray) -> str:
    """Converts a bytearray to a hex string."""
    return binascii.hexlify(bytearray(arr)).decode()


def _bytearray_to_b64_str(arr: bytearray) -> str:
    """Converts a bytearray to a base64 string."""
    return binascii.b2a_base64(bytearray(arr)).strip().decode()


class Message(object):
    """Class intended for abstracting message attributes.

    :param payload: one of the following:

       * (bytearray) including SIN and MIN bytes as first 2 in the array if not explicitly set in the call
       * (list) of integer bytes (0..255) including SIN and MIN if not specified explicitly in the call
       * (string) ASCII-HEX which includes SIN and MIN if not specified explicitly in the call
       * (string) Text which requires both SIN and MIN explictly specified in the call

    :param name: (string) optional up to 8 characters. A message name will be generated if not supplied
    :param msg_sin: integer (0..255)
    :param msg_min: integer (0..255)
    :param priority: (1=high, 4=low, 0=mobile-terminated)
    :param data_format: (optional) 1=FORMAT_TEXT, 2=FORMAT_HEX, 3=FORMAT_B64
    :param log: (optional) logger object
    :param debug: (optional) sets logging level to DEBUG

    """

    MAX_NAME_LENGTH = 8

    def __init__(self,
                 payload,
                 name: str = None,
                 msg_sin: int = None,
                 msg_min: int = None,
                 priority: int = PRIORITY_LOW,
                 data_format: int = FORMAT_HEX,
                 size: int = None,
                 logger: Logger = None,
                 debug: bool = False):
        self.log = logger or get_wrapping_logger(debug=debug)
        if name is not None:
            self.name = str(name)[0:self.MAX_NAME_LENGTH - 1]
        else:
            self.name = str(int(time()))[1:9]
            self.log.info("Message using name={}".format(self.name))
        if msg_min is not None:
            if msg_sin is None:
                raise ValueError("SIN must be specified if MIN is specified")
            elif isinstance(msg_min, int) and msg_min in range(0, 255+1):
                self.min = msg_min
                # assume that payload does not also include MIN
            else:
                self.log.warning(
                    "Invalid MIN value {} must be integer in range 0..255"
                    .format(msg_min))
        elif payload is not None:
            if isinstance(payload, bytearray):
                self.min = payload[0]
            else:
                raise ValueError(
                    "Payload must be bytearray type if MIN is not specified")
        else:
            raise ValueError("Payload cannot be None if MIN is not specified")
        if msg_sin is not None:
            if isinstance(msg_sin, int) and msg_sin in range(16, 256):
                self.sin = msg_sin
            else:
                raise ValueError(
                    "Invalid SIN value {}, must be integer in range 16..255"
                    .format(msg_sin))
        elif payload is not None:
            if isinstance(payload, bytearray):
                if payload[0] > 15:
                    self.sin = payload[0]
                    self.log.debug(
                        "Received bytearray with implied SIN={}"
                        .format(self.sin))
                    payload = payload[1:] if msg_min is None else payload[2:]
                else:
                    raise ValueError(
                        "Invalid payload, first byte (SIN) must be integer in range 16..255")
            else:
                raise ValueError(
                    "Payload must be bytearray type if SIN is not specified")
        else:
            raise ValueError("Payload cannot be None if SIN is not specified")
        self.raw_payload = bytearray(0)
        if self.sin is not None and payload is not None:
            if isinstance(payload, str):  #: TODO broken on Python2 (unicode not str)
                if data_format == FORMAT_TEXT:
                    if msg_sin is not None and msg_min is not None:
                        payload = bytearray(payload.encode())
                    else:
                        raise ValueError(
                            "Function call with text string payload must include SIN and MIN")
                elif data_format == FORMAT_HEX:
                    if _is_hex_string(payload):
                        payload = bytearray.fromhex(payload)
                    else:
                        raise ValueError(
                            "Hex format received with invalid characters")
                elif data_format == FORMAT_B64:
                    if msg_sin is not None and msg_min is not None:
                        payload = b64decode(payload)
                    else:
                        raise ValueError(
                            "Function call with base64 string payload must include SIN and MIN")
                else:
                    raise ValueError(
                        "Unrecognized data_format: {}".format(data_format))
            elif (isinstance(payload, list)
                  and all((isinstance(i, int)
                          and i in range(0, 255+1)) for i in payload)):
                payload = bytearray(payload)
            elif not isinstance(payload, bytearray):
                raise ValueError("Invalid payload {} ({}),"
                                .format(payload, type(payload))
                                + " must be text or hex string,"
                                + " integer list or bytearray")
            self.raw_payload = bytearray(payload)
            if msg_min is not None:
                self.raw_payload = bytearray([self.min]) + self.raw_payload
            if self.sin is not None:
                self.raw_payload = bytearray([self.sin]) + self.raw_payload
        self.size = len(self.raw_payload)
        if size is not None and size != self.size:
            self.log.warning(
                "Size {} passed during init does not match derived size {}"
                .format(size, self.size))
        self.priority = priority
        self.data_format = data_format
        # self.log.debug("New message created: {}".format(vars(self)))

    def data(self, data_format=FORMAT_HEX, include_min=True, include_sin=False):
        """
        Returns the data content of the message

        :param data_format: (int) 1=FORMAT_TEXT, 2=FORMAT_HEX (default), 3=FORMAT_B64
        :param include_min: (boolean) whether to include MIN byte in the data (used when not specifying MIN explicitly)
        :param include_sin: (boolean) whether to include SIN byte (not part of data for MO messages)
        :return: data as a string for submission using AT%MGRT
        """
        if len(self.raw_payload) > 0:
            if include_sin:
                if not include_min:
                    raise ValueError("Must include MIN when including SIN")
                else:
                    payload = self.raw_payload
            else:
                if include_min:
                    payload = self.raw_payload[1:]
                else:
                    payload = self.raw_payload[2:]
            if data_format == FORMAT_TEXT:
                data = '"{}"'.format(_bytearray_to_str(payload))
            elif data_format == FORMAT_HEX:
                data = _bytearray_to_hex_str(payload)
            else:
                data = _bytearray_to_b64_str(payload)
            return data
        else:
            raise ValueError("No data to return")


class MobileOriginatedMessage(Message):
    """
    Subclass of Message containing Mobile Originated (aka Return) message properties.
    Mobile-Originated state (starting=None) is represented as an attribute:

       - ``UNAVAILABLE``: 0
       - ``TX_READY``: 4
       - ``TX_SENDING``: 5
       - ``TX_COMPLETE``: 6
       - ``TX_FAILED``: 7

    :param name: (string) user identifier for the message
    :param payload: follows the structure of the Message superclass
    :param data_format: follows the structure of the Message superclass
    :param msg_sin: Service Identification Number (1st byte of payload)
    :param msg_min: Message Identification Number (2nd byte of payload)
    :param kwargs: follows the structure of the Message superclass

    """

    def __init__(self, payload, name=None, data_format=FORMAT_HEX, msg_sin=None, msg_min=None, **kwargs):
        """

        :param name: (string) user identifier for the message
        :param payload: follows the structure of the Message superclass
        :param data_format: follows the structure of the Message superclass
        :param msg_sin: Service Identification Number (1st byte of payload)
        :param msg_min: Message Identification Number (2nd byte of payload)
        :param **kwargs: follows the structure of the Message superclass

        """
        ''' TODO: remove pre-filter
        if isinstance(payload, str):
            if _is_hex_string(payload):
                payload = bytearray.fromhex(payload)
            else:
                if msg_sin is not None and msg_min is not None:
                    if data_format is None or data_format != FORMAT_TEXT:
                        payload = bytearray(payload)
                else:
                    raise ValueError(
                        "Function call with text string payload must include SIN and MIN")
        elif isinstance(payload, list) and all((isinstance(i, int) and i in range(0, 255+1)) for i in payload):
            payload = bytearray(payload)
        elif not isinstance(payload, bytearray):
            raise ValueError(
                "Invalid payload type, must be text or hex string, integer list or bytearray")
        '''
        super(MobileOriginatedMessage, self).__init__(payload=payload, name=name, msg_sin=msg_sin, msg_min=msg_min,
                                                      data_format=data_format, **kwargs)
        self.state = None


class MobileTerminatedMessage(Message):
    """
    Subclass of Message containing Mobile-Terminated (MT aka Forward) message properties.
    Initializes MT message with state = ``RX_RETRIEVED``
    MT message state represented as an attribute:

       - ``UNAVAILABLE``: 0
       - ``COMPLETE``: 2
       -  ``RETRIEVED``: 3

    :param name: (string) name assigned by the modem
    :param payload: follows the structure of the Message superclass
    :param data_format: follows the structure of the Message superclass
    :param msg_num: (string) message number assigned by the modem (unused)
    :param priority: (int) always 0 for MT messages
    :param kwargs: follows the structure of the Message superclass

    """

    def __init__(self, payload, name, data_format, msg_num=None, priority=PRIORITY_MT, **kwargs):
        """

        :param name: (string) name assigned by the modem
        :param payload: follows the structure of the Message superclass
        :param data_format: follows the structure of the Message superclass
        :param msg_num: (string) message number assigned by the modem (unused)
        :param priority: (int) always 0 for MT messages
        :param kwargs: follows the structure of the Message superclass

        """
        if data_format not in (FORMAT_TEXT, FORMAT_HEX, FORMAT_B64):
            raise ValueError(
                "Unrecognized data format: {}".format(data_format))
        super(MobileTerminatedMessage, self).__init__(payload=payload, name=name, data_format=data_format,
                                                      priority=priority, **kwargs)
        self.state = RX_RETRIEVED
        self.number = msg_num
